"""
Created on May 5, 2015

@author: Derek Wood
"""
from bi_etl.informatica.pmcmd import PMCMD
from bi_etl.scheduler.exceptions import ParameterError
from bi_etl.scheduler.task import ETLTask


class PMCMD_Task(ETLTask):
    """
    Runs Informatica Workflows
    """
    def init(self):
        """
        pre-load initialization.        
        """
        try:
            folder = self.get_parameter('folder')
        except ParameterError:
            ##  For testing purposes - Shouldn't get here
            self.log.error("PMCMD_Task didn't get folder parameter. Assuming test run")
            folder = 'MASTER'
            self.set_parameter('folder', folder)
            self.set_parameter('workflow', 'wf_TEST_Derek')
        
        self.cmd = PMCMD(config=self.config, folder= folder)

    def load(self):
        
        workflow = self.get_parameter('workflow')
        self.cmd.startworkflow(workflow)
        try:
            self.cmd.getworkflowdetails(workflow)
        except Exception:
            pass


if __name__ == '__main__':
    task = PMCMD_Task()
    task.set_parameter('folder', 'MASTER')
    task.set_parameter('workflow', 'wf_TEST_Derek')
    task.run(suppress_notifications= True)


# -*- coding: utf-8 -*-
"""
Created on Jan 5, 2016

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import typing

from bi_etl.components.row.row import Row
from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base
from bi_etl.lookups.autodisk_lookup import AutoDiskLookup
from bi_etl.lookups.disk_range_lookup import DiskRangeLookup
from bi_etl.lookups.range_lookup import RangeLookup

if typing.TYPE_CHECKING:
    from bi_etl.components.etlcomponent import ETLComponent


class AutoDiskRangeLookup(AutoDiskLookup, RangeLookup):
    """
    Automatic memory / disk lookup cache.
    
    This version divides the cache into N chunks (default is 10). If RAM usage gets beyond limits, it starts moving chunks to disk.
    Once a chunk is on disk, it stays there.
    
    TODO: For use cases where the lookup will be used in a mostly sequential fashion, it would be useful to have a version that uses ranges
    instead of a hash function. Then when find_in_cache is called on a disk segment, we could swap a different segment out and bring that 
    segment in. That's a lot more complicated. We'd also want to maintain a last used date for each segment so that if we add rows to the 
    cache, we can choose the best segment to swap to disk.
    
    Also worth considering is that if we bring a segment in from disk, it would best to keep the disk version. At that point any additions 
    to that segment would need to go to both places.

    """

    def __init__(self,
                 lookup_name: str,
                 lookup_keys: list,
                 parent_component: ETLComponent,
                 begin_date,
                 end_date,
                 config: BI_ETL_Config_Base = None,
                 use_value_cache: bool = True,
                 path=None,
                 ):
        """
        Optional parameter path controls where the data is persisted
        """
        RangeLookup.__init__(self,
                             lookup_name=lookup_name,
                             lookup_keys=lookup_keys,
                             use_value_cache=use_value_cache,
                             parent_component=parent_component,
                             begin_date=begin_date,
                             end_date=end_date,
                             config=config,
                             )
        AutoDiskLookup.__init__(self,
                                lookup_name=lookup_name,
                                lookup_keys=lookup_keys,
                                parent_component=parent_component,
                                config=config,
                                path=path,
                                begin_date=begin_date,
                                end_date=end_date,
                                use_value_cache=use_value_cache,
                                init_parent=False,  # Don't have it call the parent init because RangeLookup will have done that
                                )
        self.MemoryLookupClass = RangeLookup
        self.DiskLookupClass = DiskRangeLookup

    def cache_row(
            self,
            row: Row,
            allow_update: bool = True,
            allow_insert: bool = True,
    ):
        """
        Adds the given row to the cache for this lookup.

        Parameters
        ----------
        row: Row
            The row to cache

        allow_update: boolean
            Allow this method to update an existing row in the cache.

        allow_insert: boolean
            Allow this method to insert a new row into the cache

        Raises
        ------
        ValueError
            If allow_update is False and an already existing row (lookup key) is passed in.

        """
        AutoDiskLookup.cache_row(self, row, allow_update=allow_update)

    def find_in_cache(self, row, **kwargs):
        return RangeLookup.find_in_cache(self, row=row, **kwargs)



"""
Created on Apr 2, 2015
"""
import os
import re
import sys
from pathlib import Path
from typing import *

from openpyxl import load_workbook, Workbook
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.table import Table as WorksheetTable, TableStyleInfo

from bi_etl.components.row.row import Row
from bi_etl.components.xlsx_reader import XLSXReader
from bi_etl.scheduler.task import ETLTask
from bi_etl.statistics import Statistics

__all__ = ['XLSXWriter']


class XLSXWriter(XLSXReader):
    """
    XLSXWriter will write rows to a Microsoft Excel XLSX formatted workbook.
    
    Parameters
    ----------
    task: ETLTask
        The  instance to register in (if not None)
    
    file_name: str
        The file_name to parse as xlsx.
        
    logical_name: str
        The logical name of this source. Used for log messages.

    write_only: bool
        Should we use the faster write only mode ?
        (only supports brand-new files not adding data to an existing file)

    Attributes
    ----------
    column_names: list
        The names to use for columns
        
    header_row: int
        The sheet row to read headers from. Default = 1.
    
    start_row: int
        The first row to parse for data. Default = header_row + 1 
    
    workbook: :class:`openpyxl.workbook.workbook.Workbook`
        The workbook that was opened.
        
    log_first_row : boolean
        Should we log progress on the first row read. *Only applies if Table is used as a source.*
        (inherited from ETLComponent)
        
    max_rows : int, optional
        The maximum number of rows to read. *Only applies if Table is used as a source.*
        (inherited from ETLComponent)
        
    primary_key: list
        The name of the primary key column(s). Only impacts trace messages.  Default=None.
        (inherited from ETLComponent)
    
    progress_frequency: int
        How often (in seconds) to output progress messages. None for no progress messages.
        (inherited from ETLComponent)
    
    progress_message: str
        The progress message to print. Default is ``"{logical_name} row # {row_number}"``.
        Note ``logical_name`` and ``row_number`` subs.
        (inherited from ETLComponent)
        
    restkey: str
        Column name to catch extra long rows (more columns than we have column names)
        when reading values (extra values).
        
    restval: str
        The value to put in columns that are in the column_names but 
        not present in a given row when reading (missing values).
    """
    def __init__(self,
                 task: Optional[ETLTask],
                 file_name: Union[str, Path],
                 logical_name: str = None,
                 write_only: bool = True,
                 **kwargs
                 ):
        super().__init__(
            task=task,
            file_name=file_name,
            logical_name=logical_name,
        )
        self.headers_written = False
        self.write_only = write_only
        self._insert_cnt = 0
        self._insert_cnt_this_sheet = 0

        # Matches invalid XML1.0 unicode characters, like control characters:
        # http://www.w3.org/TR/2006/REC-xml-20060816/#charsets
        # http://stackoverflow.com/questions/1707890/fast-way-to-filter-illegal-xml-unicode-chars-in-python

        _illegal_unichrs = [(0x00, 0x08), (0x0B, 0x0C), (0x0E, 0x1F),
                            (0x7F, 0x84), (0x86, 0x9F),
                            (0xFDD0, 0xFDDF), (0xFFFE, 0xFFFF)]
        if sys.maxunicode >= 0x10000:  # not narrow build
            _illegal_unichrs.extend([(0x1FFFE, 0x1FFFF), (0x2FFFE, 0x2FFFF),
                                     (0x3FFFE, 0x3FFFF), (0x4FFFE, 0x4FFFF),
                                     (0x5FFFE, 0x5FFFF), (0x6FFFE, 0x6FFFF),
                                     (0x7FFFE, 0x7FFFF), (0x8FFFE, 0x8FFFF),
                                     (0x9FFFE, 0x9FFFF), (0xAFFFE, 0xAFFFF),
                                     (0xBFFFE, 0xBFFFF), (0xCFFFE, 0xCFFFF),
                                     (0xDFFFE, 0xDFFFF), (0xEFFFE, 0xEFFFF),
                                     (0xFFFFE, 0xFFFFF), (0x10FFFE, 0x10FFFF)])

        _illegal_ranges = ["%s-%s" % (chr(low), chr(high))
                           for (low, high) in _illegal_unichrs]
        self._illegal_xml_chars_RE = re.compile(u'[%s]' % u''.join(_illegal_ranges))

        # Should be the last call of every init
        self.set_kwattrs(**kwargs)

    def __repr__(self):
        return "XLSXWriter({})".format(self.logical_name)

    @property
    def rows_inserted(self):
        return self._insert_cnt

    @property
    def rows_inserted_this_sheet(self):
        return self._insert_cnt_this_sheet

    @property
    def workbook(self):
        if self._workbook is None:
            if self.write_only:
                self._workbook = Workbook(write_only=True)
            else:
                if os.path.isfile(self.file_name):
                    self._workbook = load_workbook(filename=self.file_name, read_only=False)
                else:
                    self._workbook = Workbook()
                    del self._workbook['Sheet']
        return self._workbook

    def set_active_worksheet_by_name(self, sheet_name):
        self._active_worksheet_name = sheet_name
        if sheet_name not in self.workbook:
            self._active_worksheet = self.workbook.create_sheet(sheet_name)
            self._insert_cnt_this_sheet = 0
            if self.headers_written:
                self._column_names = None
                self.headers_written = False
        else:
            self._active_worksheet = self.workbook[sheet_name]
            self._column_names = None
            self._insert_cnt_this_sheet = 0

    def set_active_worksheet_by_number(self, sheet_number: int):
        """
        Change to an existing worksheet based on the sheet number.
        """
        try:
            super().set_active_worksheet_by_number(sheet_number)
        except ValueError as e:
            raise ValueError(f"{e}. Use set_active_worksheet_by_name to create new sheets.")

    def _obtain_column_names(self):
        raise ValueError(f'Column names must be explicitly set on {self}')

    def set_columns_and_widths(self, columns_dict: Dict[str, float]):
        """
        Set the column names and widths at the same time.
        See :py:meth:`set_widths`.
        """
        self.column_names = columns_dict.keys()
        self.set_widths(columns_dict.values())

    def write_header(self):
        """
        Write the header row.
        """
        if self.column_names is None:
            raise ValueError("insert called before column_names set (or possibly column_names needs to be set after set_active_worksheet_by_name call.")
        self.active_worksheet.append(self.column_names)
        self.headers_written = True

    def insert_row(
        self,
        source_row: Row,  # Must be a single row
        additional_insert_values: dict = None,
        stat_name: str = 'insert',
        parent_stats: Statistics = None,
    ) -> Row:
        """
        Inserts a row into the database (batching rows as batch_size)

        Parameters
        ----------
        source_row:
            The row with values to insert
        additional_insert_values:
            Values to add / override in the row before inserting.
        stat_name:
            Name of this step for the ETLTask statistics.
        parent_stats:
            Optional Statistics object to nest this steps statistics in.
            Default is to place statistics in the ETLTask level statistics.

        Returns
        -------
        new_row
        """
        stats = self.get_stats_entry(stat_name, parent_stats=parent_stats)
        stats.timer.start()

        # row_values = [source_row[col] for col in self.column_names]

        if not self.headers_written:
            self.write_header()

        assert len(source_row) > 0, f"Empty row passed into {self} insert"

        values = []
        for column_name in self.column_names:
            if additional_insert_values and column_name in additional_insert_values:
                col_value = additional_insert_values[column_name]
            else:
                try:
                    col_value = source_row[column_name]
                except KeyError:
                    col_value = None
            if isinstance(col_value, str):
                col_value = self._illegal_xml_chars_RE.sub('\\?', col_value)
            values.append(col_value)

        new_row = self.row_object(iteration_header=self.full_iteration_header)
        new_row.update_from_values(values)

        assert len(values) > 0, f"No values with column names from {self} found in {source_row}"

        self.active_worksheet.append(values)

        self._insert_cnt += 1
        self._insert_cnt_this_sheet += 1

        stats.timer.stop()

        return new_row

    def set_widths(self, column_widths: Iterable[float]):
        """
        Set the column widths in the xlsx for the currently active worksheet.
        """
        for i, column_width in enumerate(column_widths):
            self.active_worksheet.column_dimensions[get_column_letter(i + 1)].width = column_width

    def make_table_from_inserted_data(self, style_name: str = 'TableStyleMedium2'):
        """
        Format the newly inserted worksheet data as a table.

        Parameters
        ----------

        style_name:
            The name of the Excel style to apply to the table.
        """
        if self._insert_cnt_this_sheet == 0:
            return
        last_col_letter = get_column_letter(len(self.column_names))

        table_ref = f"A1:{last_col_letter}{self._insert_cnt_this_sheet + 1}"
        self.log.debug(f"Adding table to worksheet {self.active_worksheet_name} for range {table_ref}")

        table = WorksheetTable(
            displayName=f"Tbl_{self.active_worksheet_name}",
            ref=table_ref
        )
        style = TableStyleInfo(
            name=style_name,
            showFirstColumn=False,
            showLastColumn=False,
            showRowStripes=True,
            showColumnStripes=False
        )
        table.tableStyleInfo = style

        # noinspection PyProtectedMember
        table._initialise_columns()
        for column, value in zip(table.tableColumns, self.column_names):
            column.name = value
        self.active_worksheet.add_table(table)

    def insert(
        self,
        source_row: Union[Row, list],  # Could also be a whole list of rows
        parent_stats: Statistics = None,
        **kwargs
    ):
        """
        Insert a row or list of rows in the table.

        Parameters
        ----------
        source_row:
            Row(s) to insert
        parent_stats:
            Optional Statistics object to nest this steps statistics in.
            Default is to place statistics in the ETLTask level statistics.
        """

        if isinstance(source_row, list):
            for row in source_row:
                self.insert_row(
                    row,
                    parent_stats=parent_stats,
                    **kwargs
                )
        else:
            self.insert_row(
                 source_row,
                 parent_stats=parent_stats,
                 **kwargs
             )
            
    def close(self, error: bool = False):
        """
        Close the xlsx file, saving first if ``error`` is false.

        Parameters
        ----------
        error:
            Did we run into an error during processing?
            Errors cause a rollback, which skips the save of the file.
        """
        if self.has_workbook_init():
            if not error:
                self.workbook.save(filename=self.file_name)
            else:
                self.log.info(f"{self}.close not saving due to error")
            self.workbook.close()
        super().close(error=error)



# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import gzip
import io
import json
import os.path
import os.path
from typing import *
from datetime import datetime
from tempfile import TemporaryDirectory

from bi_etl.bulk_loaders.redshift_s3_base import RedShiftS3Base
from bi_etl.bulk_loaders.s3_bulk_load_config import S3_Bulk_Loader_Config
from bi_etl.timer import Timer

if TYPE_CHECKING:
    from bi_etl.components.table import Table
    from bi_etl.scheduler.task import ETLTask


class RedShiftS3JSONBulk(RedShiftS3Base):
    def __init__(self,
                 config: S3_Bulk_Loader_Config,
                 ):
        super().__init__(
            config=config,
        )

    @property
    def needs_all_columns(self):
        return False

    def get_copy_sql(
            self,
            s3_source_path: str,
            table_to_load: str,
            file_compression: str = '',
            analyze_compression: str = None,
            options: str = '',
    ):
        analyze_compression = analyze_compression or self.analyze_compression
        if analyze_compression:
            options += f' COMPUPDATE {self.analyze_compression} '

        return f"""\
                COPY {table_to_load} FROM 's3://{self.s3_bucket_name}/{s3_source_path}'                      
                     credentials 'aws_access_key_id={self.s3_user_id};aws_secret_access_key={self.s3_password}'
                     JSON 'auto'
                     {file_compression}  
                     {options}
               """

    @staticmethod
    def json_serializer(value):
        if isinstance(value, datetime):
            return str(value)
        else:
            # raise ValueError(f'No json_serializer support for {repr(value)}')
            str(value)

    def load_from_iterator(
            self,
            iterator: Iterator,
            table_object: Table,
            table_to_load: str = None,
            perform_rename: bool = False,
            progress_frequency: int = 10,
            analyze_compression: str = None,
            parent_task: Optional[ETLTask] = None,
    ) -> int:
        row_count = 0
        with TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            local_files = []
            zip_pool = []
            text_wrapper_pool = []

            writer_pool_size = self.s3_files_to_generate

            for file_number in range(writer_pool_size):
                filepath = os.path.join(temp_dir, f'data_{file_number}.json.gz')
                local_files.append(filepath)
                zip_file = gzip.open(filepath, 'wb')
                text_wrapper = io.TextIOWrapper(zip_file, encoding='utf-8')
                text_wrapper_pool.append(text_wrapper)
                zip_pool.append(zip_file)

            progress_timer = Timer()
            for row_number, row in enumerate(iterator):
                row_count += 1
                text_wrapper = text_wrapper_pool[row_number % writer_pool_size]
                text_wrapper.write(json.dumps(row.as_dict, default=self.json_serializer))
                text_wrapper.write("\n")
                if progress_frequency is not None:
                    # noinspection PyTypeChecker
                    if 0 < progress_frequency < progress_timer.seconds_elapsed:
                        self.log.info(f"Wrote row {row_number:,}")
                        progress_timer.reset()

            for text_wrapper in text_wrapper_pool:
                text_wrapper.close()

            for zip_file in zip_pool:
                zip_file.close()

            if row_count > 0:
                self.load_from_files(
                    local_files,
                    file_compression='GZIP',
                    table_object=table_object,
                    table_to_load=table_to_load,
                    perform_rename=perform_rename,
                    analyze_compression=analyze_compression,
                )
            else:
                self.log.info(f"{self} had nothing to do with 0 rows found")

            return row_count



"""
Created on Mar 20, 2015

@author: Derek Wood
"""
import sys
import typing
from collections import OrderedDict

OrderedDict = dict if sys.version_info >= (3, 6) else OrderedDict
from typing import List

from bi_etl.timer import Timer
from bi_etl.utility import dict_to_str


class Statistics(object):
    """
    classdocs
    """

    def __init__(self, stats_id: str, parent: 'Statistics' = None, print_start_stop_times: bool = True):
        """
        Constructor
        """
        self.stats_id = stats_id
        self._stats_data = OrderedDict()
        self.parent_path = None
        if parent is not None:
            self.parent_path = parent.path
            parent[stats_id] = self
        self._timer = Timer(start_running=False)

        self.print_start_stop_times = print_start_stop_times

        # Add place-holders to the OrderedDict so that these timer stats are listed in keys or iterators
        # Needs to be at the end of __init__ so the place-holders don't set it to True
        self.used = False

    def __str__(self):
        return self.stats_id

    def __repr__(self):
        return dict_to_str(self)

    # Allow a Statistics object to follow the ETLTask interface as used by bi_etl.scheduler.task.ETLTask#statistics
    @property
    def statistics(self):
        return self

    @property
    def path(self) -> List[str]:
        if self.parent_path is not None:
            return self.parent_path + [self.stats_id]
        else:
            return [self.stats_id]

    def get_unique_stats_id(self, stats_id) -> str:
        id_nbr = 1
        base_stats_id = stats_id
        while stats_id in self:
            id_nbr += 1
            stats_id = '{} {}'.format(base_stats_id, id_nbr)
        return stats_id

    @property
    def timer(self):
        self.used = True
        return self._timer

    def __getitem__(self, key):
        if key == 'start time':
            return self.timer.first_start_time
        elif key == 'stop  time':
            return self.timer.stop_time
        elif key == 'seconds elapsed':
            return self.timer.seconds_elapsed
        elif key in self._stats_data:
            return self._stats_data[key]
        else:
            # Makes stats['new stat'] += x possible (although we do have add_to_stat)
            return 0

    def __setitem__(self, key, value):
        # Don't count assignment of 0 or None as used
        if value:
            self.used = True
        self._stats_data[key] = value
        if isinstance(value, Statistics):
            value.parent_path = self.path
            value.stats_id = key

    def add_to_stat(self, key, increment):
        if key in self:
            self[key] += increment
        else:
            self[key] = increment

    def ensure_exists(self, key, initial_value=0):
        if key not in self:
            self[key] = initial_value

    def iteritems(self):
        # Needs to be based on keys() because we force extra special keys
        for k in list(self.keys()):
            yield k, self[k]

    def items(self):
        # Needs to be based on keys() because we force extra special keys
        return list(self.iteritems())

    def keys(self):
        key_list = list()
        if self.timer.start_time is not None:
            if self.print_start_stop_times:
                key_list.append('start time')
                key_list.append('stop  time')
            key_list.append('seconds elapsed')
        key_list += list(self._stats_data.keys())
        return key_list

    def values(self):
        # Needs to be based on keys() because we force extra special keys
        values_data = []
        for k in list(self.keys()):
            values_data.append(self[k])
        return values_data

    def update(self, other):
        return self._stats_data.update(other)

    def merge(self, other):
        for stats_key in other.keys():
            value = other[stats_key]
            if isinstance(value, int):
                self.add_to_stat(stats_key, value)
            elif (isinstance(value, dict)
                  or isinstance(value, Statistics)):
                if self[stats_key] == 0:
                    self[stats_key] = value
                elif isinstance(self[stats_key], Statistics):
                    self[stats_key].merge(value)
                elif isinstance(self[stats_key], OrderedDict):
                    self[stats_key] = Statistics(stats_id=stats_key,
                                                 parent=self)
                    self[stats_key].merge(value)
                else:
                    raise ValueError(
                        "Can't merge {path}:{key} type {type} "
                        "with {opath}:{okey} type {otype}".format(
                            path=self.path,
                            key=stats_key,
                            type=type(self[stats_key]),
                            opath=other.path,
                            okey=stats_key,
                            otype=type(value),
                        )
                    )

    def __contains__(self, key):
        return self._stats_data.__contains__(key)

    def get(self, key, default=None):
        if key in self:
            return self[key]
        else:
            return default

    def __len__(self):
        return len(self._stats_data)

    def __delitem__(self, key):
        self._stats_data.__delitem__(key)

    def __iter__(self):
        for k in list(self.keys()):
            yield self[k]

    @staticmethod
    def format_statistics(container):
        return dict_to_str(container,
                           show_list_item_number=False,
                           show_type=False,
                           show_length=False,
                           indent_per_level=4,
                           type_formats={
                               int:   ',',
                               float: '.3f',
                           },
                           )

    @staticmethod
    def find_item(obj: typing.Union[str, dict, list], key: str):
        if isinstance(obj, str):
            return None
        elif hasattr(obj, 'values'):
            # If this key matches the target key, return it's value
            if key in obj:
                return obj[key]
            # Otherwise recursively check values
            for v in list(obj.values()):
                item = Statistics.find_item(v, key)
                if item is not None:
                    return item
                # Otherwise keep iterating keys
        elif isinstance(obj, list):
            for v in obj:
                item = Statistics.find_item(v, key)
                if item is not None:
                    return item



"""
Created on May 5, 2015

@author: Derek Wood
"""

from config_wrangler.config_templates.credentials import Credentials
from pydantic import FilePath


class PMCMDConfig(Credentials):
    INFA_HOME: str = None
    INFA_DOMAINS_FILE: FilePath = None
    USER_SECURITY_DOMAIN: str = None
    REPOSITORY: str
    SERVICE: str
    DOMAIN: str
    DEFAULT_FOLDER: str
    DEFAULT_PARMFILE: str = None
    DEFAULT_LOCALPARAMFILE: str = None
    OSPROFILE: str = None
    RUN_VIA_CMD: bool = False



# -*- coding: utf-8 -*-
"""
Created on Dec 23, 2015

@author: Derek Wood
"""
from bi_etl.scheduler.task import ETLTask


class CheckPointTask(ETLTask):
    """
    A checkpoint virtual task in the dependency tree that exists just to allow for one set of tasks to wait for another whole set of tasks to complete.

    @staticmethod depends_on() should be overridden.

    TaskA--\             /--- Task D
    TaskB----Checkpoint------ Task E
    TaskC--/
    """

    def load(self):
        pass

    def run(self,
            suppress_notifications=None,
            parent_to_child=None,
            child_to_parent=None,
            handle_exceptions=True,
            ):
        pass



"""
Created on Mar 2, 2015

@author: Derek Wood
"""
import logging
from datetime import datetime, date
from typing import Dict, Any, List, Mapping, Tuple

import sqlalchemy

from bi_etl.components.row.row import Row


class StatementQueue(object):
    """
    A queue of pending SQLAlchemyy statements
    """

    def __init__(self, execute_with_binds=True):
        """
        Constructor
        """
        self.statements: Dict[Any, sqlalchemy.sql.base.Executable] = dict()
        self.statement_values: Dict[Any, List[Mapping]] = dict()
        self.row_count: int = 0
        self.execute_with_binds: bool = execute_with_binds
        self.row_limit: int = 1000
        # Batch size = 	65,536 * Network Packet Size
        # Network Packet Size is the size of the tabular data stream (TDS) packets used to communicate between
        # applications and the relational Database Engine. The default packet size is 4 KB, and is controlled by the
        # network packet size configuration option.
        # We set the size soft limit as below this because
        # 1) It's a soft limit, we might go higher
        # 2) It doesn't count the INSERT INTO clause part (e.g. INSERT INTO perf_test (c1,c2,dt,i1,f1,d1) VALUES )
        self.size_soft_limit = 60000 * 4096
        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        
    def __len__(self):
        return self.row_count
    
    def get_statement_by_key(self, key):
        return self.statements.get(key)
    
    def add_statement(self, key, stmt):
        self.statements[key] = stmt

    @staticmethod
    def encapsulate_value(value) -> str:
        # Encapsulate the input for SQL use (add ' etc)
        if isinstance(value, str):
            return "'" + value.replace("'", "''") + "'"
        # Note we need to check datetime before date because a date passes both isinstance
        elif isinstance(value, datetime):
            if value.microsecond == 0:
                return "'" + value.isoformat() + "'"
            else:
                return f"'{value:%Y-%m-%d %H:%M:%S}.{int(value.microsecond / 1000)}{value:%z}'"
        elif isinstance(value, date):
            return "'" + value.isoformat() + "'"
        elif value is None:
            return 'Null'
        else:
            return str(value)

    @staticmethod
    def encapsulate_row_value_list(value_list: list) -> str:
        return ','.join([StatementQueue.encapsulate_value(v) for v in value_list])

    def values_str_list(self, rows, row_limit: int = None, size_soft_limit: int = None):
        if row_limit is None:
            row_limit = self.row_limit
        if size_soft_limit is None:
            size_soft_limit = self.size_soft_limit
        values_str_list = []
        values_str = []
        line_counter = 0
        size = 0
        for i in rows:
            limit_reached = False
            if line_counter >= row_limit:
                limit_reached = True
                self.log.debug(f'insert stmt row limit {row_limit:,} reached')
            elif size_soft_limit is not None and size > size_soft_limit:
                limit_reached = True
                self.log.debug(f'insert stmt size limit {size_soft_limit:,} reached or exceeded with {size:,} for {line_counter:,} lines')
            if limit_reached:
                values_str_list.append(",".join(values_str))
                values_str = []
                line_counter = 0
                size = 0
            row_str = "(" + StatementQueue.encapsulate_row_value_list(i) + ")"
            values_str.append(row_str)
            size += len(row_str) + 1  # One extra for comma separator
            line_counter += 1
        # Add the final list
        values_str_list.append(",".join(values_str))
        return values_str_list

    def execute(self, connection) -> int:
        rows_affected = 0
        if self.row_count > 0:
            for stmtKey in self.statements.keys():
                stmt = self.statements[stmtKey]
                values = self.statement_values[stmtKey]
                if len(values) > 0:
                    if self.execute_with_binds:
                        result = connection.execute(stmt, values)
                        if result.rowcount == -1:
                            rows_affected += len(values)
                        else:
                            rows_affected += result.rowcount
                    else:
                        # cursor = connection.engine.raw_connection().cursor()
                        # for values_str in self.values_str_list(values):
                        #     cursor.execute(stmt.format(values=values_str))
                        #     rows_affected += cursor.rowcount
                        for values_str in self.values_str_list(values):
                            # Treat stmt as a string
                            # noinspection PyUnresolvedReferences
                            result = connection.execute(stmt.format(values=values_str))
                            if result.rowcount == -1:
                                rows_affected += len(values)
                            else:
                                rows_affected += result.rowcount
                self.statement_values[stmtKey].clear()
            self.row_count = 0
        return rows_affected
    
    def iter_single_statements(self) -> Tuple[sqlalchemy.sql.base.Executable, Row]:
        if self.row_count > 0:
            for stmtKey in self.statements.keys():
                stmt = self.statements[stmtKey]
                values = self.statement_values[stmtKey]
                for row in values:
                    yield stmt, row

    def execute_singly(self, connection) -> int:
        rows_affected = 0
        for row_num, (stmt, row) in enumerate(self.iter_single_statements()):
            try:
                if self.execute_with_binds:
                    result = connection.execute(stmt, row)
                    rows_affected += result.rowcount
                else:
                    # cursor = connection.engine.raw_connection().cursor()
                    # for values_str in StatementQueue.values_str_list(row, limit=1):
                    #     cursor.execute(stmt.format(values=values_str))
                    #     rows_affected += cursor.rowcount
                    for values_str in self.values_str_list(rows=[row], row_limit=1):
                        result = connection.execute(stmt.format(values=values_str))
                        rows_affected += result.rowcount
            except Exception as e:
                try:
                    vals = row.str_formatted()
                except AttributeError:
                    vals = row
                msg = f"Error {e} with row {row_num} stmt {stmt} stmt_values {vals}"
                self.log.error(msg)
                try:
                    orig = e.orig
                except AttributeError:
                    orig = None
                raise sqlalchemy.exc.StatementError(msg, str(stmt), vals, orig)
        return rows_affected
        
    def append_values_by_key(self, key, values):
        values_list = self.statement_values.get(key)
        if values_list is None:
            values_list = list()
            self.statement_values[key] = values_list
        values_list.append(values)
        self.row_count += 1    



"""
Created on Apr 23, 2015

@author: Derek Wood
"""


class TaskStopRequested(Exception):
    pass

class ParameterError(Exception):
    pass

class WorkflowFinished(Exception):
    pass

class CircularDependency(Exception):
    def __init__(self, circular_list):
        self.circular_list = circular_list
    
    def __repr__(self):
        return "CircularDependency(circular_list=\n{}".format(self.circular_list)
    
class DependencyDeeperThanLimit(Exception):
    def __init__(self, limit, maxed_list):
        self.limit = limit
        self.maxed_list = maxed_list
    
    def __repr__(self):
        return "DependencyDeeperThanLimit(limit={}, maxed_list={}".format(self.limit, self.maxed_list)


"""
Created on Apr 14, 2015

@author: Derek Wood
"""
import codecs
import multiprocessing
import textwrap
import time
import traceback
import warnings
from datetime import datetime, timedelta
from queue import Empty

from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.orm.exc import DetachedInstanceError

import bi_etl.scheduler
import bi_etl.scheduler.dependent_reasons as dependent_reasons
from bi_etl.conversions import replace_tilda
from bi_etl.scheduler.exceptions import CircularDependency
from bi_etl.scheduler.exceptions import DependencyDeeperThanLimit
from bi_etl.scheduler.messages import ChildSetDisplayName
from bi_etl.scheduler.messages import (ChildStatusUpdate,
                                       ChildRunRequested,
                                       ChildRunOK,
                                       ChildNotOccupyingCPU,
                                       ChildOccupyingCPU,
                                       SummaryMessage
                                       )
from bi_etl.scheduler.models import (ETL_Tasks,
                                     ETL_Task_Log,
                                     ETL_Task_Stats,
                                     ETL_Task_Dependency
                                     )
from bi_etl.scheduler.scheduler_etl_jobs.etl_task_status_cd import ETL_Task_Status_CD
from bi_etl.scheduler.scheduler_interface import SchedulerInterface
from bi_etl.scheduler.task import Status, TaskStopRequested, ETLTask
from bi_etl.statistics import Statistics
from bi_etl.utility import dict_to_pairs


#pylint: disable=too-many-instance-attributes, too-many-public-methods, too-many-nested-blocks 
#pylint: disable=too-many-statements, too-many-branches, too-many-arguments, too-many-locals
#pylint: disable=invalid-name

# Ignore long lines for now
#pylint: disable=line-too-long

class Scheduler(SchedulerInterface):
    """
    Full-scheduler object with memory of active jobs.
    """

    MAX_DEPENDENTS_DEPTH = 100

    def __init__(self):
        super(Scheduler, self).__init__(log_name='Scheduler', allow_create= True)
        
        # Register the replace_tilda method of handling unicode errors 
        codecs.register_error('replace_tilda', replace_tilda)
        
        # Use the spawn method to get a clean python instance for each ETL Job
        multiprocessing.set_start_method('spawn')  # @UndefinedVariable
        self.trace = self.config.getboolean('Scheduler', 'trace', fallback=False)

        self.log.info("Init full Scheduler instance")

        # TODO: Setup to monitor config file and reload if it changes

        # Only used by the full scheduler so we import here
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import psutil
        self.psutil = psutil

        self.etl_task_by_id = dict()
        self.etl_task_by_name = dict()
        maximum_concurrent_tasks = self.config.getint_('Scheduler',
                                                       'maximum_concurrent_tasks',
                                                       fallback=None
                                                       )
        if maximum_concurrent_tasks is None:
            try:
                self.maximum_concurrent_tasks = multiprocessing.cpu_count()  # @UndefinedVariable
                msg = """
                      Setting maximum_concurrent_tasks to CPU count of {}.
                      Provide Scheduler.maximum_concurrent_tasks in config if something else is desired.
                      """
                msg = textwrap.dedent(msg)
                self.log.warning(msg.format(self.maximum_concurrent_tasks))
            except NotImplementedError as e:
                self.log.warning(e)
                self.maximum_concurrent_tasks = 1
                msg = """
                      Defaulting to {} concurrent tasks. 
                      Provide maximum_concurrent_tasks in config if something else is desired.
                      """
                msg = textwrap.dedent(msg)
                msg = msg.format(self.maximum_concurrent_tasks)
                self.log.warning(msg)
        else:
            self.maximum_concurrent_tasks = maximum_concurrent_tasks
        self.log.info("maximum_concurrent_tasks = {}".format(maximum_concurrent_tasks))
        self.tasks_occupying_cpu = set()
        self.log_flush_interval = 10
        self.process_check_interval = self.config.getint('Scheduler', 'process_check_interval', fallback=5)
        self.seconds_to_keep_finished_tasks = 60
        self.days_to_keep_run_logs = 30
        self.days_to_keep_large_parameters = 7
        self.min_large_parameter_size = 1024  # Parameters over 1 KB are considered large
        self.last_table_maintenance_run = None
        self.log.debug("Finished Init full Scheduler instance")
        self._new_tasks_query = None

        heartbeat_age_timedelta = self.get_heartbeat_age_timedelta()
        if (heartbeat_age_timedelta is not None 
            and heartbeat_age_timedelta < timedelta(seconds= self.process_check_interval * 2)
            ):
            raise RuntimeError('Another scheduler appears to be active!')

    def _create_etl_task_object_from_task_rec(self, task_rec):
        task_id = task_rec.task_id
        if not isinstance(task_id, int):
            task_id = int(task_id)
        parent_task_id = task_rec.parent_task_id
        root_task_id = task_rec.root_task_id
        parent_task = None
        if parent_task_id is not None:
            parent_task = self.get_task_by_id(task_rec.parent_task_id)
            if parent_task is None:
                self.log.warning("Task {} {} started with parent id {} that is not known".format(task_id, task_rec.modulename, parent_task_id))
            else:
                # Check parent status
                if parent_task.status == Status.succeeded:
                    self.add_log_message(parent_task, 'Child started status changing to ancestors_running')
                    self.set_task_status(parent_task, Status.ancestors_running)
                elif parent_task.status.is_terminated() or parent_task.status.is_termination_pending():
                    task_rec.Status = Status.stop_requested
                elif parent_task.status == Status.failed:
                    #pylint: disable= redefined-variable-type
                    task_rec.Status = Status.stopped

                if root_task_id != parent_task_id and root_task_id != parent_task.root_task_id:
                    self.log.warning("Creating task {task_id} listed as has having parent_task_id={parent_task_id} "
                                     "which has root_task_id={root_task_id}...".format(
                                        task_id = task_id,
                                        parent_task_id = parent_task_id,
                                        root_task_id = parent_task.root_task_id,
                                        )
                                     )
                    self.log.warning("...however, parent database record has root_task_id={}".format(root_task_id))
                    self.log.warning("...correcting the task database row")
                    task_rec.root_task_id = parent_task.root_task_id
                    root_task_id = parent_task.root_task_id
        if root_task_id is not None:
            root_task = self.get_task_by_id(root_task_id)
            if parent_task is None:
                parent_task = root_task
            if root_task is None:
                self.log.warning("Task {} {} started with root id {} that is not known".format(task_id, task_rec.modulename, root_task_id))
            else:
                if root_task.task_rec.Status.is_finished():
                    self.set_task_status(root_task, Status.ancestors_running)
        try:
            # Support finding the instance using either module alone (which could include an embedded classname) or module + classname
            if task_rec.classname:                
                instance_name = task_rec.modulename + '.' + task_rec.classname
            else:
                instance_name = task_rec.modulename
            
            ETLTask_Class = self.find_etl_class_instance(partial_module_name= instance_name)
        except ValueError as e:
            self.log.exception(e)
            self._write_log_message(task_rec.task_id, repr(e))
            task_rec.summary_message = repr(e)
            task_rec.Status = Status.failed
            self.session.commit()
            return None
        
        try:
            etl_task = ETLTask_Class(task_id= task_id,
                                     parent_task_id=parent_task_id,
                                     root_task_id=root_task_id,
                                     scheduler= self,
                                     task_rec = task_rec,
                                     config= self.config,
                                     )
        # Scheduler doesn't die no matter what!
        except Exception as e:  # pylint: disable=broad-except
            self.log.exception(e)
            self._write_log_message(task_rec.task_id, repr(e))
            task_rec.summary_message = repr(e)
            task_rec.Status = Status.failed
            self.session.commit()
            return None
        
        self.log.debug("Task {} object created in memory".format(repr(etl_task)))
        # Make extra sure ETLTask stored the task_ids properly
        assert etl_task.task_id == task_id
        assert etl_task.parent_task_id == parent_task_id
        assert etl_task.root_task_id == root_task_id

        return etl_task

    def get_task_by_id(self, etl_task_id):
        if not isinstance(etl_task_id, int):
            etl_task_id = int(etl_task_id)
        # Check active tasks in memory
        if etl_task_id in self.etl_task_by_id:
            return self.etl_task_by_id[etl_task_id]
        else:
            # Otherwise try to get it from the database
            task_rec = self.get_task_record(etl_task_id)
            return self._create_etl_task_object_from_task_rec(task_rec)

    def get_etl_tasks_by_name(self, etl_task_name):
        if etl_task_name in self.etl_task_by_name:
            return self.etl_task_by_name[etl_task_name]
        else:
            return []

    def add_task_to_active(self, etl_task):
        assert isinstance(etl_task, ETLTask)
        if not isinstance(etl_task.task_id, int):
            etl_task.task_id = int(etl_task.task_id)
        self.etl_task_by_id[etl_task.task_id] = etl_task
        task_name = etl_task.name
        if task_name not in self.etl_task_by_name:
            self.etl_task_by_name[task_name] = dict()
        self.etl_task_by_name[task_name][etl_task.task_id] = etl_task

    def remove_active_task_record(self, etl_task):
        # Last chance to flush message queue
        self.flush_log_messages(etl_task)
        try:
            etl_task_name = etl_task.name

            if not hasattr(etl_task, 'thread_running'):
                # Stopped before it was started
                pass
            elif etl_task.thread_running:
                self.log.warning("remove_active_task_record called with Task {} but has thread_running = True".format(repr(etl_task)))
            else:
                self.log.debug("Removing completed task {} from active tasks".format(etl_task))
                self.session.flush()
                try:

                    self.session.expunge(etl_task.task_rec)
                except AttributeError:
                    pass
                except InvalidRequestError:
                    pass
                del self.etl_task_by_id[etl_task.task_id]
                del self.etl_task_by_name[etl_task_name][etl_task.task_id]
                # Remove children
                for child_task in etl_task.children:
                    self.remove_active_task_record(child_task)
        except KeyError:
            self.log.error(traceback.format_exc())
            self.log.error("ETL Task {} was not in etl_task_by_name".format(repr(etl_task)))

    def set_task_status(self, etl_task, new_status):
        self.log.debug("set_task_status task_id={} new_status={}".format(etl_task.task_id, new_status))
        self.add_log_message(etl_task, "set_task_status new_status={}".format(new_status))

        # Set status on both the object and the task_rec
        # TODO: Clean that up
        etl_task.status = new_status

        if new_status.is_finished():
            etl_task.task_rec.finished_date = datetime.now()
        #self.log.debug("Task row = {}".format(dict_to_str(etl_task.task_rec)))
        self.session.flush()
        self.session.commit()
        # Check if parent wants child status updates
        if etl_task.parent_task is not None:
            if etl_task.parent_task.needs_to_get_child_statuses():
                self.send_task_message(etl_task.parent_task, ChildStatusUpdate(etl_task.task_id, new_status))
        # Check if root wants ancestor status updates
        if etl_task.root_task is not None and etl_task.root_task != etl_task.parent_task:
            if etl_task.root_task.needs_to_get_ancestor_statuses():
                self.send_task_message(etl_task.root_task, ChildStatusUpdate(etl_task.task_id, new_status))

    def send_task_message(self, etl_task, message):
        try:
            etl_task.parent_to_child.put(message)
        except AttributeError:
            self.log.warning("send_task_message called on task {} with no parent_to_child".format(etl_task))

    def set_task_summary_message(self, etl_task, summary_message, commit = False, from_client= False):
        """
        set the tasks summary message.
        Note: Since this is often called on conjunction with set_task_status, we default to no commit here.
        """
        self.add_log_message(etl_task, msg = summary_message, allow_duplicates= False)
        self.log.debug("set_task_summary_message task={} message={}".format(etl_task, summary_message))
        if etl_task.summary_message_from_client and not from_client:
            self.log.debug("Not storing summary message because it would override a client message")
        else:
            etl_task.task_rec.summary_message = str(summary_message)[:4000]
            if commit:
                self.session.commit()

    def append_task_summary_message(self, etl_task, summary_message, commit = True):
        """
        Append to the tasks summary message.
        """
        self.set_task_summary_message(etl_task, etl_task.task_rec.summary_message + ' ' + summary_message)
        if commit:
            self.session.commit()

    def check_for_cpu(self, etl_task):
        """
        Checks for available CPU, if available it STARTS THE TASK and returns True.
        If not it returns False
        """
        if len(self.tasks_occupying_cpu) < self.maximum_concurrent_tasks:
            self.run_task(etl_task)
            return True
        else:
            return False

    @staticmethod
    def _depth_first_dependent_search(start, target, limit, current_depth=0, call_tree=None):
        if current_depth >= limit:
            raise DependencyDeeperThanLimit(limit, call_tree)

        # Make sure we are dealing with the ETL_Tasks task record from the model and not the ETL_Task object
        if isinstance(start, ETLTask):
            start = start.task_rec

        if isinstance(target, ETLTask):
            target = target.task_rec

        if call_tree is None:
            call_tree = list()
        call_tree.append(start)

        if start.task_id == target.task_id:
            return call_tree

        for u in start.dependencies:
            #if u.dependent_reason == dependent_reasons.Predecessor:
            # TODO: Track the u.dependent_reason in the call_tree
            if Scheduler._depth_first_dependent_search(u.dependent_on_task, target, limit, current_depth+1, call_tree):
                return call_tree

        # Remove ourselves from the master call tree
        call_tree.remove(start)
        return None

    def add_dependency(self, etl_task, dependent_task, dependent_reason):
        done = False
        assert isinstance(etl_task, ETLTask), 'add_dependency: etl_task is not an ETLTask but is a {}'.format(type(etl_task))
        assert isinstance(dependent_task, ETLTask), 'add_dependency: dependent_task is not an ETLTask but is a {}'.format(type(dependent_task))
        if dependent_task.task_id == ETLTask.task_id:
            self.add_log_message(etl_task, 'Not recording dependency on {} since that is this task'.format(repr(dependent_task)))
            return
        # Don't depend on cross root tasks that are already dependent on something in our root
        if dependent_task.task_rec.root_task_id != etl_task.task_rec.root_task_id:
            for dep_dep_record in dependent_task.task_rec.dependencies:
                dep_dep_task_rec = dep_dep_record.dependent_on_task
                if dep_dep_task_rec.root_task_id == etl_task.task_rec.root_task_id:
                    self.add_log_message(etl_task, 'Not recording dependency on {} since it is waiting on our root'.format(repr(dependent_task)))
                    done = True

        if not done:
            for exist_dep in etl_task.task_rec.dependencies:
                if exist_dep.dependent_on_task_id == dependent_task.task_id:
                    if exist_dep.current_blocking_flag == 'N':
                        self.add_log_message(etl_task, 'Re-blocking dependency on {}'.format(repr(dependent_task)))
                        exist_dep.current_blocking_flag = 'Y'
                    else:
                        self.add_log_message(etl_task, 'Already dependent on {} blocking_flag = {}'.format(repr(dependent_task), exist_dep.current_blocking_flag))
                    done = True

        if not done:
            # Make sure dependent_task isn't also directly or indirectly dependent on etl_task
            loop = Scheduler._depth_first_dependent_search(start= dependent_task.task_rec,
                                                           target= etl_task.task_rec,
                                                           limit= Scheduler.MAX_DEPENDENTS_DEPTH
                                                           )
            if loop is not None:
                raise CircularDependency(loop)
            else:
                self.add_log_message(etl_task, 'Adding dependency on {}'.format(repr(dependent_task)))
                dep = ETL_Task_Dependency(etl_task.task_rec, dependent_task.task_rec, str(dependent_reason))
                self.session.add(dep)

    def check_dependencies(self, etl_task):
        #pylint: disable=protected-access
        
        assert isinstance(etl_task, ETLTask), "check_dependencies needs an ETLTask got a {}".format(type(etl_task))
        
        if not etl_task._logged_dependencies:
            self.add_log_message(etl_task, 'depends_on = {}'.format(etl_task.depends_on()))
            self.add_log_message(etl_task, 'normalized_dependents_set = {}'.format(etl_task.normalized_dependents_set))
            self.add_log_message(etl_task, 'mutually_exclusive_with_set = {}'.format(etl_task.mutually_exclusive_with_set))
            etl_task._logged_dependencies = True

        dependencies_met = True

        # Scan existing dependencies to check if they are still relevant
        for dep in etl_task.task_rec.dependencies:
            if dep.current_blocking_flag == 'Y':
                dep_task = self.get_task_by_id(dep.dependent_on_task_id)
                dependent_reason_class = dependent_reasons.getFromString(dep.dependent_reason)
                if dependent_reason_class is None:
                    self.add_log_message(
                        etl_task,
                        f"Unable to process dependent_reason {dep}. "
                        f"Assuming it will get re-added. dep_task= {repr(dep_task)}"
                    )
                    dep.current_blocking_flag = 'N'
                elif dependent_reason_class.release_wait_on_status(dep_task.status):
                    self.add_log_message(etl_task, 'No longer blocked by {}'.format(repr(dep_task)))
                    dep.current_blocking_flag = 'N'
                else:
                    if self.trace:
                        self.log.debug(f"{repr(etl_task)} still blocked by {dep} {repr(dep_task)}")
                    dependencies_met = False
            elif dep.current_blocking_flag != 'N':
                self.log.warning(f'{dep} current_blocking_flag has unexpected value of {dep.current_blocking_flag}')

        # Check for new dependencies

        # Check jobs that we depend on to make sure they are not actively running or failed
        # For example given jobs A-> B -> C
        # If A is running then B should not start
        # If A failed, then B should fail
        for dep_name in etl_task.normalized_dependents_set:
            # Check if that task name is active (running, or waiting)
            dep_obj_dict = self.get_etl_tasks_by_name(dep_name)
            if dep_obj_dict is not None and len(dep_obj_dict) > 0:
                # Scan the entire list as dependent objects
                for dep_task_id in dep_obj_dict:
                    dep_task = dep_obj_dict[dep_task_id]
                    assert isinstance(dep_task, ETLTask)

                    if dependent_reasons.Predecessor.wait_on_status(dep_task.task_rec.Status):
                        try:
                            self.add_dependency(etl_task, dep_task, dependent_reasons.Predecessor)
                            self.log.debug("Task {etl_task} has to wait because dependent "
                                           "job {dep} has not is_finished yet.".format(
                                                etl_task= repr(etl_task),
                                                dep=repr(dep_task)
                            ))
                            dependencies_met = False
                        except CircularDependency as error:
                            # Somehow that task is waiting for us, so we'll skip waiting on it
                            self.add_log_message(etl_task, 'Did not add dependency on {} due to {}'.format(
                                repr(dep_task),
                                error
                            ))
                    else:  # Failed
                        if etl_task.root_task_id is not None and dep_task.root_task_id == etl_task.root_task_id:
                            # If our predecessor within the same root failed, cancel this job
                            if dep_task.status.value < 0 or dep_task.status.is_termination_pending():
                                msg = "Predecessor {} failed with status {}".format(
                                    repr(dep_task),
                                    repr(dep_task.status)
                                )
                                self.log.debug("Task {etl_task} {msg}".format(etl_task=repr(etl_task), msg=msg))
                                self.add_log_message(etl_task, msg + "\n", allow_duplicates=False)
                                self.set_task_summary_message(etl_task, "Predecessor {} failed".format(
                                    dep_task.task_id
                                ))
                                self.set_task_status(etl_task, Status.cancelled)
                                return
                        else:
                            # Different root task instance failed (any tasks with root of None are considered different)
                            pass
            else:
                self.add_log_message(etl_task, "Predecessor {} is not registered with scheduler".format(dep_name))

        # Check to see if another instance of this task is running
        # Check if that task name is active (running, or waiting)
        # TODO: Have a way that tasks can indicate if it's OK for another instance to run at the same time.
        if dependencies_met:
            for exclusive_etl_task in etl_task.mutually_exclusive_with_set:
                exclusive_etl_task_obj_dict = self.get_etl_tasks_by_name(exclusive_etl_task)
                for exclusive_etl_task_instance_task_id in exclusive_etl_task_obj_dict:
                    exclusive_etl_task_instance = exclusive_etl_task_obj_dict[exclusive_etl_task_instance_task_id]
                    if exclusive_etl_task_instance.task_id != etl_task.task_id:
                        # Check if that task also lists us as Mutually exclusive
                        if etl_task.name in exclusive_etl_task_instance.mutually_exclusive_with_set:
                            # If so treat like another instance of self.
                            # We wait only on those before us in line (based on Task ID)
                            if dependent_reasons.OtherSelf.wait_on_status(exclusive_etl_task_instance.status):
                                if etl_task.task_id > exclusive_etl_task_instance.task_id:
                                    try:
                                        msg = "Waiting for other waiting instance {} with status {}".format(
                                            exclusive_etl_task_instance.task_id,
                                            exclusive_etl_task_instance.status
                                        )
                                        self.add_dependency(etl_task,
                                                            exclusive_etl_task_instance,
                                                            dependent_reasons.OtherSelf)
                                        self.add_log_message(etl_task, msg + "\n", allow_duplicates=False)
                                        self.log.debug("Task {etl_task} has to wait because another instance {other}"
                                                       " is also waiting and arrived first.".format(
                                                         etl_task= repr(etl_task),
                                                         other=repr(exclusive_etl_task_instance)
                                        ))
                                        dependencies_met = False
                                    except CircularDependency as error:
                                        # Somehow that task is waiting for us, so we'll skip waiting on it
                                        self.add_log_message(etl_task, 'Did not add dependency on {} due to {}'.format(
                                            repr(exclusive_etl_task_instance), error
                                        ))
                        else:  # Other task isn't directly listing us a Mutually Exclusive (which it probably should)
                            # Do we add ourselves to exclusive_etl_task_instance.mutually_exclusive_with_set ?
                            if dependent_reasons.MutuallyExclusive.wait_on_status(exclusive_etl_task_instance.status):
                                self.log.debug("Task {etl_task} has to wait because an instance of a mutually exclusive"
                                               " task {other} has not is_finished yet.".format(
                                                  etl_task= repr(etl_task),
                                                  other=repr(exclusive_etl_task_instance)
                                ))
                                try:
                                    msg = "Waiting for an instance of a mutually exclusive" \
                                          " task {} (and maybe others)".format(exclusive_etl_task_instance.task_id)
                                    self.add_log_message(etl_task, msg + "\n", allow_duplicates=False)
                                    self.add_dependency(etl_task,
                                                        exclusive_etl_task_instance,
                                                        dependent_reasons.MutuallyExclusive)
                                    dependencies_met = False
                                except CircularDependency as error:
                                    # Somehow that task is waiting for us, so we'll skip waiting on it
                                    self.add_log_message(etl_task,
                                                         'Did not add dependency on {} due to {}'.format(
                                                             repr(exclusive_etl_task_instance),
                                                             error
                                                         ))

        # Check to make sure there are no running jobs that list this job as a dependency
        # That would make this a source of that job and so we don't want to start running until that job finishes
        # For example given jobs A -> B -> C
        # If B is running then A should not start.  If C is running A should be OK to start.
        if dependencies_met:
            for running_etl_task_id in self.etl_task_by_id:
                running_etl_task = self.get_task_by_id(running_etl_task_id)
                if dependent_reasons.Follower.wait_on_status(running_etl_task.status):
                    for dep_name in running_etl_task.normalized_dependents_set:
                        if dep_name == etl_task.name:
                            try:
                                self.add_dependency(etl_task, running_etl_task, dependent_reasons.Follower)
                                dependencies_met = False
                                self.log.debug(
                                    "Task {etl_task} has to wait because following job {flw} is still active.".format(
                                        etl_task= etl_task,
                                        flw=running_etl_task
                                        )
                                    )
                            except CircularDependency as error:
                                self.log.warning(error)
                                self.add_log_message(etl_task,
                                                     'Did not add dependency on {} due to {}'.format(
                                                         repr(running_etl_task),
                                                         error
                                                     ))
        if dependencies_met:
            # check if parent has needs_to_ok_child_runs() == True which sets waiting_for_workflow flag
            if etl_task.waiting_for_workflow:
                if not etl_task.parent_task:
                    self.add_log_message(etl_task, 'waiting_for_workflow: Cannot wait for non-existing parent!')
                    self.set_task_status(etl_task, Status.failed)
                elif not etl_task.parent_task.thread_running:
                    self.add_log_message(etl_task,
                                         'waiting_for_workflow: Parent not running to approve us! (parent={})'.format(
                                             repr(etl_task.parent_task)
                                         ))
                    self.set_task_status(etl_task, Status.failed)
                else:
                    # Send a message requesting etl_tasks run (should have already been sent)
                    self.send_task_message(etl_task.parent_task, message=ChildRunRequested(etl_task.task_id))
                    self.set_task_status(etl_task, Status.waiting_for_workflow)
            else:
                # Check for cpu right away (rather than waiting for the next pass)
                if not self.check_for_cpu(etl_task):
                    # If check_for_cpu didn't start the task, set it to wait for CPU
                    self.set_task_status(etl_task, Status.waiting_for_cpu)
        else:
            if etl_task.status == Status.new:
                self.set_task_status(etl_task, Status.waiting_for_dependencies)
            # This should be non-empty here, but lets check
            if len(etl_task.task_rec.dependencies) > 0:
                waiting_for_list = [str(d) for d in etl_task.task_rec.dependencies if d.current_blocking_flag == 'Y']
                waiting_for = "Waiting for {}".format(', '.join(waiting_for_list))
                self.set_task_summary_message(etl_task, waiting_for)
            else:
                self.set_task_summary_message(etl_task, "Dependencies not met")

    def check_for_new_task_commands(self):
        """
        Looks in the database repository for new task commands:
          * new task: Creates an ETLTask instance, and checks it's predesssors.
          * stop_requested: Sends a stop signal to the task
          * kill_requested: Sends s SIGKILL to terminate the task
        """

        #TODO: For new tasks this method of reading ETL_Tasks works. However, for stop requests it can fail.
        #      If the web-interface writes a stop request right as the scheduler updates the status message,
        #      the stop status will get overwritten by the status message update.
        #      Proposed solution: Create a scheduler message queue table that can be used to send messages
        #      to the scheduler. In environments where it's possible, the web-interface could also use
        #      a sockets interface to communicate directly with the scheduler.

        try:
            if self._new_tasks_query is None:
                valid_status_codes = [s.value for s in [Status.new, Status.stop_requested, Status.kill_requested]]
                query = self.session.query(ETL_Tasks).filter(ETL_Tasks.status_id.in_(valid_status_codes) )
                query = query.filter( ETL_Tasks.scheduler_id == self.scheduler_id )
                query = query.order_by(ETL_Tasks.task_id)
                self._new_tasks_query = query
            else:
                query = self._new_tasks_query
            for task_rec in query.all():
                if task_rec.Status == Status.new:
                    self.log.info("Registering new task {r.modulename} id {r.task_id} parent_task_id "
                                  "{r.parent_task_id} root_task_id {r.root_task_id}".format(
                                     r=task_rec
                    ))
                    etl_task = self._create_etl_task_object_from_task_rec(task_rec)
                    if etl_task is not None:
                        self.add_task_to_active(etl_task)
                        self.check_dependencies(etl_task)
                    else:
                        self.log.info("--Failed registering new task {r.modulename} id {r.task_id} "
                                      "parent_task_id {r.parent_task_id} root_task_id {r.root_task_id}".format(
                                          r=task_rec
                        ))
                elif task_rec.Status == Status.stop_requested:
                    if task_rec.task_id in self.etl_task_by_id:
                        try:
                            etl_task = self.etl_task_by_id[task_rec.task_id]
                            self.stop_task(etl_task)
                        # Scheduler doesn't die no matter what!
                        except Exception:  #pylint: disable=broad-except
                            task_rec.summary_message = traceback.format_exc()
                            self.session.commit()
                    else:
                        # it's not running, just change it's listing to stopped
                        self.log.warning("Got request to stop task that is not active. task_id={}".format(task_rec.task_id))
                        task_rec.Status = Status.stopped
                        self.session.commit()
                elif task_rec.Status == Status.kill_requested:
                    if task_rec.task_id in self.etl_task_by_id:
                        try:
                            etl_task = self.etl_task_by_id[task_rec.task_id]
                            self.kill_task(etl_task)
                        # Scheduler doesn't die no matter what!
                        except Exception:  #pylint: disable=broad-except
                            task_rec.summary_message = traceback.format_exc()
                            self.session.commit()
                    else:
                        # it's not running, just change it's listing to stopped
                        self.log.warning("Got request to kill task that is not active. task_id={}".format(
                            task_rec.task_id
                        ))
                        task_rec.Status = Status.stopped
                        self.session.commit()
                else:
                    self.log.error("Got task with un-expected status in "
                                   "check_for_new_task_commands. task_id={} status={}".format(
                                       task_rec.task_id,
                                       task_rec.Status
                    ))
        # Scheduler doesn't die no matter what!
        except Exception as e:  #pylint: disable=broad-except
            self.log.error(traceback.format_exc())
            self.log.error(repr(e))

    def run_task(self, etl_task):
        self.log.info("Starting task {}".format(etl_task))
        self.add_log_message(etl_task, 'Starting process')
        try:
            parent_to_child = multiprocessing.Queue()  # @UndefinedVariable
            child_to_parent = multiprocessing.Queue()  # @UndefinedVariable
            # Run method to avoid serializing the task itself
            process = multiprocessing.Process(target=bi_etl.scheduler.task.run_task,  # @UndefinedVariable
                                              kwargs={
                                                      'task_name': etl_task.name,
                                                      'task_id': etl_task.task_id,
                                                      'parent_task_id': etl_task.parent_task_id,
                                                      'root_task_id': etl_task.root_task_id,
                                                      'parent_to_child': parent_to_child,
                                                      'child_to_parent': child_to_parent,
                                                      'scheduler': self,
                                                     }
                           )
            process.start()
            etl_task.thread_running = True
            # After starting add the Queues to our copy of the etl_task object
            etl_task.parent_to_child = parent_to_child
            etl_task.child_to_parent = child_to_parent
            etl_task.process = process

            etl_task.task_rec.started_date = datetime.now()

            self.tasks_occupying_cpu.add(etl_task.task_id)
            self.log.debug('tasks_occupying_cpu = {}'.format(self.tasks_occupying_cpu))
            etl_task.task_rec.pid = process.pid
            self.set_task_summary_message(etl_task, None)
            self.set_task_status(etl_task, Status.running)
        # Scheduler doesn't die no matter what!
        except Exception as e:  #pylint: disable=broad-except
            self.add_log_message(etl_task,  traceback.format_exc() )
            self.set_task_summary_message(etl_task, repr(e))
            self.set_task_status(etl_task, Status.failed)

    def stop_task(self, etl_task):
        if etl_task.status.is_alive():
            self.set_task_summary_message(etl_task, "Stopping")
            self.log.debug("Sending stop request to {}".format(repr(etl_task)))
            self.send_task_message(etl_task, message='stop')
            self.set_task_status(etl_task, Status.stop_signal_sent)
        elif not etl_task.status.is_finished():
            self.set_task_status(etl_task, Status.stopped)

    def kill_task(self, etl_task):
        if etl_task.status.is_alive():
            self.set_task_summary_message(etl_task, "Killing")
            etl_task.process.terminate()
            if etl_task.status != Status.kill_signal_sent:
                self.set_task_status(etl_task, Status.kill_signal_sent)
        elif not etl_task.status.is_finished():
            self.set_task_status(etl_task, Status.stopped)

    def _write_log_message(self, task_id, msg, time_now=None):
        if msg is not None:
            if time_now is None:
                time_now = datetime.now()

            msg_rec = ETL_Task_Log()
            msg_rec.task_id = task_id
            msg_rec.log_entry_ts = time_now
            # TODO: despite having the database column support uncode, cx_oracle is getting
            # errors when we send unicode through. So we'll trip it out for now.
            msg = msg.encode('ascii', 'replace_tilda').decode('ascii')
            msg_rec.log_entry = msg
             
            self.session.add(msg_rec)

    def flush_log_messages(self, etl_task, time_now= None):

        if len(etl_task.pending_log_msgs) > 0:
            self.log.debug("flush_log_messages task_id={}".format(etl_task.task_id))
            if time_now is None:
                time_now = datetime.now()
            # Make sure log messages are at least 1 second apart to avoid PK violation
            if (etl_task.last_log_msg_time is not None
                and ((time_now - etl_task.last_log_msg_time) < timedelta(seconds=1))
               ):
                    time_now = etl_task.last_log_msg_time + timedelta(seconds = 1)
                    # Set the new last time
                    etl_task.last_log_msg_time = time_now

            if self.trace:
                self.log.debug("flush_log_messages task_id={} add msg".format(etl_task.task_id))

            log_entry = ''.join(etl_task.pending_log_msgs)
            self._write_log_message(etl_task.task_id, msg=log_entry, time_now=time_now)

            if self.trace:
                self.log.debug("flush_log_messages task_id={} commit".format(etl_task.task_id))            
            self.session.commit()
            del etl_task.pending_log_msgs[:]
            etl_task.last_log_msg_time = time_now
            if self.trace:
                self.log.debug("flush_log_messages task_id={} done".format(etl_task.task_id))

    def check_if_log_message_flush_needed(self, etl_task):
        time_now = datetime.now()
        if etl_task.last_log_msg_time is None:
            # Set the initial value to when we got called first,
            # this will delay the first log storage until log_flush_interval
            etl_task.last_log_msg_time = time_now
        elif (time_now - etl_task.last_log_msg_time) > timedelta(seconds=self.log_flush_interval):
            self.flush_log_messages(etl_task, time_now)

    def add_log_message(self, etl_task, msg, allow_duplicates=True):
        if msg is None:
            return

        if self.trace:
            self.log.debug("add_log_message task_id={} msg=---{}---".format(etl_task.task_id, msg.strip()))
        msg_ok = True
        if msg != etl_task.last_log_msg:
            etl_task.last_log_msg = msg
        else:
            if not allow_duplicates:
                msg_ok = False
        if msg_ok:
            if not isinstance(msg, str):
                msg = str(msg)
            # Add an ending new line if it's not there
            if not msg.endswith('\n'):
                msg += '\n'
            etl_task.pending_log_msgs.append(msg)
        else:
            if self.trace:
                self.log.debug("add_log_message task_id={} msg=---{}--- not saved (duplicate)".format(etl_task.task_id, msg.strip()))

        #self.check_if_log_message_flush_needed(etl_task)

    def process_statistics(self, etl_task, stats):
        task_rec = etl_task.task_rec
        for stat_name, stat_value in dict_to_pairs(stats):
            stat_name = str(stat_name)
            stat_value = str(stat_value)
            if stat_name in task_rec.stats:
                if task_rec.stats[stat_name] != stat_value:
                    task_rec.stats[stat_name] = ETL_Task_Stats(stat_name, stat_value)
            else:
                task_rec.stats[stat_name] = ETL_Task_Stats(stat_name, stat_value)
        self.session.commit()

    @staticmethod
    def _is_reloaded_instance(msg, class_obj):
        """
        Like isinstance but it should deal with modules that have been reloaded.
        Unlike isinstance it will NOT handle cases where msg inherits from class_obj
        """
        return str(type(msg)) == str(class_obj)

    def read_child_messages(self, etl_task):
        is_reloaded_instance = self._is_reloaded_instance
        q = etl_task.child_to_parent
        # Check that we actually have a queue
        if q is None:
            return
        try:
            while True:
                msg = q.get_nowait()
                if is_reloaded_instance(msg, Statistics):
                    self.process_statistics(etl_task, msg)
                elif isinstance(msg, str):  # basestring won't be affected by reload, so use the normal isinstance
                    self.add_log_message(etl_task, msg)
                elif is_reloaded_instance(msg, TaskStopRequested):
                    self.set_task_summary_message(etl_task, "Stopped")
                    self.set_task_status(etl_task, Status.stopped)
                elif isinstance(msg, Exception):  # Exception won't be affected by reload, so use the normal isinstance
                    self.set_task_summary_message(etl_task, repr(msg), from_client= True)
                elif is_reloaded_instance(msg, Status):
                    if etl_task.status not in [Status.stopped, Status.stop_signal_sent]:
                        self.log.debug("read_child_messages setting {} status to {}".format(etl_task, msg))
                        self.set_task_status(etl_task, msg)
                elif is_reloaded_instance(msg, ChildSetDisplayName):
                    child_task_id = msg.child_task_id
                    child_task = self.get_task_by_id(child_task_id)
                    child_task.display_name = msg.display_name
                elif is_reloaded_instance(msg, ChildRunOK):
                    child_task_id = msg.child_task_id
                    child_task = self.get_task_by_id(child_task_id)
                    if child_task is None:
                        self.log.error("Got ChildRunOK for invalid child task_id {}".format(child_task_id))
                    else:
                        child_task.waiting_for_workflow = False
                        msg = "Got ChildRunOK for {}, clearing waiting_for_workflow flag".format(child_task)
                        self.add_log_message(child_task, msg)
                        if child_task.status == Status.waiting_for_workflow:
                            self.set_task_status(child_task, Status.waiting_for_dependencies)
                        else:
                            msg = "After ChildRunOK on {}, leaving status as {}".format(child_task, child_task.status)
                            self.add_log_message(child_task, msg)
                elif is_reloaded_instance(msg, ChildNotOccupyingCPU):
                    child_task_id = msg.child_task_id
                    if child_task_id in self.tasks_occupying_cpu:
                        self.tasks_occupying_cpu.remove(child_task_id)
                    else:
                        child_task = self.get_task_by_id(child_task_id)
                        msg = "{} Sent ChildNotOccupyingCPU but it wasn't occupying CPU".format(child_task)
                        self.add_log_message(child_task, msg)
                        self.log.warning(msg)
                elif is_reloaded_instance(msg, ChildOccupyingCPU):
                    child_task_id = msg.child_task_id
                    if child_task_id not in self.tasks_occupying_cpu:
                        self.tasks_occupying_cpu.add(child_task_id)
                    else:
                        child_task = self.get_task_by_id(child_task_id)
                        msg = "{} Sent ChildOccupyingCPU but it already was occupying CPU".format(child_task)
                        self.add_log_message(child_task, msg)
                        self.log.warning(msg)
                elif is_reloaded_instance(msg, SummaryMessage):
                    if msg.append:
                        self.append_task_summary_message(etl_task, msg.summary_message)
                    else:
                        self.set_task_summary_message(etl_task, msg.summary_message, from_client= True)
                else:
                    self.log.warning("Got unexpected message from client: {}".format(repr(msg)))
        except Empty:
            pass

    def check_if_done(self, etl_task):
        try:
            self.log.debug("Calling is_alive: {}".format(repr(etl_task)))
            if hasattr(etl_task, 'process') and etl_task.process.is_alive():
                if etl_task.status.is_finished():
                    self.log.warning("Task {} is completed but is_alive is still true.".format(repr(etl_task)))
                    try:
                        self.log.warning("...Process={}".format(etl_task.process))
                        self.log.warning("...Process PID={}".format(etl_task.process.pid))
                    except AttributeError:
                        pass
            else:
                if hasattr(etl_task, 'thread_running') and etl_task.thread_running:
                    self.log.debug("Calling join: {}".format(repr(etl_task)))
                    self.add_log_message(etl_task, "Cleaning up thread")
                    etl_task.process.join(10)
                    if etl_task.task_id in self.tasks_occupying_cpu:
                        self.tasks_occupying_cpu.remove(etl_task.task_id)
                        self.log.debug('tasks_occupying_cpu = {}'.format(self.tasks_occupying_cpu))
                    etl_task.thread_running = False
                    if etl_task.process.exitcode != 0:
                        self.log.warning("etl_task {} had exit code {}".format(repr(etl_task), etl_task.process.exitcode))
                        self.add_log_message(etl_task, "exit code {}".format( etl_task.process.exitcode))
                    if etl_task.status == Status.succeeded:
                        self.set_task_summary_message(etl_task, "Done", commit=True)

                if etl_task.status in [Status.running, Status.succeeded, Status.ancestors_running]:
                    children_done = True
                    children_status = None
                    for child_task in etl_task.children:
                        if child_task.status != child_task.task_rec.Status:
                            self.log.error('child_task.status {} != child_task.task_rec.Status {}'.format(child_task.status, child_task.task_rec.Status))
                            self.add_log_message(etl_task,'ERROR child_task.status {} != child_task.task_rec.Status {}'.format(child_task.status, child_task.task_rec.Status))
                            child_task.status = child_task.task_rec.Status
                        if not child_task.status.is_finished():
                            children_done = False
                            self.log.debug("check_if_done etl_task {} waiting for child {} in status {}".format(repr(etl_task), repr(child_task), child_task.status))
                            self.add_log_message(etl_task, "Waiting for child {} in status {}".format(repr(child_task), child_task.status))
                            break
                        else:  # Get the "worst" child status... the lowest.
                            self.add_log_message(etl_task, "Collecting status for child {} in status {}".format(repr(child_task), child_task.status))
                            if children_status is None:
                                children_status = child_task.status
                                self.log.debug("check_if_done etl_task {} first child status {}".format(repr(etl_task), children_status))
                            elif child_task.status < children_status:
                                children_status = child_task.status
                                self.log.debug("check_if_done etl_task {} new low child status {}".format(repr(etl_task), children_status))
                    if children_done:
                        # Set is_finished time
                        self.add_log_message(etl_task, "Task and all ancestors is_finished")
                        self.set_task_summary_message(etl_task, 'Done')
                        self.log.debug("check_if_done: {} not is_alive and children_done. "
                                       "etl_task.status={} children_status={}".format(
                                          repr(etl_task),
                                          etl_task.status,
                                          children_status
                        ))
                        if children_status is not None:
                            if children_status == Status.failed:
                                self.set_task_status(etl_task, Status.ancestor_failed)
                            else:
                                self.set_task_status(etl_task, children_status)
                        else:
                            # Double check for child messages that might update our status
                            self.read_child_messages(etl_task)
                            # If still running then task must have failed
                            if etl_task.status == Status.running:
                                self.set_task_status(etl_task, Status.failed)
                        self.log.debug("check_if_done: {} chilren_done etl_task.status={}".format(repr(etl_task), etl_task.status))
                        self.add_log_message(etl_task, "Final status = {}".format(etl_task.status))
                    else:  # children not done
                        if etl_task.status != Status.ancestors_running:
                            # From above we know status is in
                            # [Status.running, Status.succeeded, Status.ancestors_running]
                            self.set_task_status(etl_task, Status.ancestors_running)
                elif etl_task.status == Status.failed:
                    pass
                elif etl_task.status == Status.stop_signal_sent:
                    self.set_task_status(etl_task, Status.stopped)
                elif etl_task.status == Status.stopped:
                    pass
                elif etl_task.status == Status.kill_signal_sent:
                    self.set_task_status(etl_task, Status.killed)
                else:
                    self.log.warning("Unexpected status {} in check_if_done for task {}".format(
                        etl_task.status,
                        repr(etl_task)
                    ))
                    self.set_task_status(etl_task, Status.failed)

        except AttributeError as e:
            self.log.error("check_if_done: {}".format(repr(e)))

    def check_for_previous_instance_tasks(self):
        """
        Looks in the database repository for tasks from a previous instance.
        """
        running_status_codes = list()
        for s in Status:
            if not s.is_finished() and not s == Status.new:
                running_status_codes.append(s.value)
        query = self.session.query(ETL_Tasks).filter(ETL_Tasks.status_id.in_(running_status_codes) )
        query = query.filter( ETL_Tasks.scheduler_id == self.scheduler_id )
        query = query.order_by(ETL_Tasks.task_id)
        our_process = self.psutil.Process()
        wait_pids = list()
        for task_rec in query.all():
            self.log.info("Found task {task_rec.task_id} {task_rec.modulename} "
                          "for previous scheduler instance listed as active".format(
                              task_rec=task_rec
            ))
            pid = task_rec.pid
            if pid and self.psutil.pid_exists(pid):
                try:
                    process = self.psutil.Process(pid)
                    if (process.is_running()
                        and (process.exe() == our_process.exe())
                        and (process.gids() == our_process.gids())
                       ):
                        self.log.info("  {task_rec.task_id} {task_rec.modulename} is still running "
                                      "as pid {task_rec.pid}".format(task_rec=task_rec))
                        wait_pids.append(pid)
                        process.terminate()
                        task_rec.summary_message = "Killed (active task PID) on scheduler restart. " \
                                                   "Prior status was {}".format(task_rec.Status)
                    else:
                        task_rec.summary_message = "Killed (PID didn't match) on scheduler restart. " \
                                                   "Prior status was {}".format(task_rec.Status)
                except self.psutil.NoSuchProcess:
                    pass
            else:
                task_rec.summary_message = "Killed (PID not active) on scheduler restart. " \
                                           "Prior status was {}".format(task_rec.Status)
            task_rec.Status = Status.killed
            self.log.info("  {task_rec.task_id} {task_rec.modulename} "
                          "new status = {task_rec.status_id}".format(task_rec=task_rec))

        self.session.commit()
        if len(wait_pids) > 0:
            self.log.info("Waiting for previous schedulers tasks to end.")
            self.psutil.wait_procs(wait_pids, timeout=30)
            self.log.info("Done waiting for previous schedulers tasks to end.")

    def delete_old_run_logs(self):
        query = self.session.query(ETL_Tasks)
        query = query.filter(ETL_Tasks.submitted_date <
                             (datetime.now()-timedelta(days=self.days_to_keep_run_logs) )
                             )
        query.delete()

    def purge_old_lager_parameters(self):
        query = self.session.query(ETL_Tasks)
        query = query.filter(ETL_Tasks.submitted_date <
                             (datetime.now()-timedelta(days=self.days_to_keep_large_parameters) )
                             )
        for etl_task_rec in query:
            for param in etl_task_rec.parameters:
                if len(param.param_value) > self.min_large_parameter_size:
                    param.param_value = 'PURGED'

    def check_for_required_table_maintenance(self):
        # Delete etl_task records from the database after retention period X days
        if (self.last_table_maintenance_run is None
            or (datetime.now() - self.last_table_maintenance_run) > timedelta(hours=24)
           ):
            self.log.info("Performing table cleanup")
            #logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)
            #logging.getLogger('sqlalchemy.engine.base.Engine').setLevel(logging.DEBUG)
            self.delete_old_run_logs()
            self.purge_old_lager_parameters()
            self.session.commit()
            self.last_table_maintenance_run = datetime.now()

    def start_monitoring(self):
        #logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)
        #logging.getLogger('sqlalchemy.engine.base.Engine').setLevel(logging.DEBUG)
        #logging.getLogger('sqlalchemy.orm').setLevel(logging.DEBUG)

        # Check for tasks that are listed as running, or waiting from the last run of the scheduler
        self.log.info("check_for_previous_instance_tasks")
        self.check_for_previous_instance_tasks()
        
        self.scan_etl_classes()

        self.log.info("Running status code upsert task")
        self.add_task_by_class(ETL_Task_Status_CD)

        self.log.info("Starting monitoring for new jobs")

        while True:
            try:
                self.check_for_new_task_commands()

                list_to_remove = list()

                for etl_task_id in self.etl_task_by_id:
                    etl_task = self.etl_task_by_id[etl_task_id]
                    #self.log.debug("Checking on task {}".format(repr(etl_task)))
                    self.session.refresh(etl_task.task_rec)        
                    status = etl_task.status
                    if not status.is_finished():
                        self.log.debug("Checking task {} with status {}".format(repr(etl_task), status))

                    if etl_task.thread_running or etl_task.status.is_termination_pending():
                        # Read any messages in the queue for this task
                        self.read_child_messages(etl_task)
                        self.check_if_done(etl_task)
                    elif status.is_finished():
                        # Check if we can remove from memory (active tasks)
                        if etl_task.root_task_id is None:
                            if etl_task.task_rec.finished_date is not None:
                                if ((datetime.now() - etl_task.task_rec.finished_date)
                                    > timedelta(minutes=self.seconds_to_keep_finished_tasks)
                                   ):
                                    list_to_remove.append(etl_task)
                            else:
                                self.log.warning("Task {} has is_finished status but no finished_date.".format(
                                    repr(etl_task)
                                ))
                    elif status == Status.ancestors_running:
                        self.check_if_done(etl_task)
                    elif status == Status.waiting_for_cpu:
                        self.check_for_cpu(etl_task)
                    elif status == Status.waiting_for_dependencies:
                        self.check_dependencies(etl_task)
                    else:
                        msg = "Unexpected status {} in monitoring loop for task {}".format(status, etl_task)
                        self.log.warning(msg)
                    self.flush_log_messages(etl_task)
                # End for loop of etl_task's

                etl_task = None

                for etl_task in list_to_remove:
                    self.remove_active_task_record(etl_task)

                # Update heart beat time in etl_schedulers table
                self.heatbeat_now()

                self.check_for_required_table_maintenance()

                self.session.commit()
            except (CircularDependency, DependencyDeeperThanLimit) as e:
                self.add_log_message(etl_task, repr(e))
                self.set_task_summary_message(etl_task, "Invalid dependency. See log.")
                self.set_task_status(etl_task, Status.failed)
                self.flush_log_messages(etl_task)
            # Scheduler doesn't die no matter what!
            except DetachedInstanceError:
                raise SystemExit
            except Exception as e:  #pylint: disable=broad-except
                self.log.error(traceback.format_exc())
                self.log.error(repr(e))
                try:
                    self.session.commit()                
                except Exception:  #pylint: disable=broad-except
                    # OK, things are more serious here. Rollback?
                    # We might just not have a database connection anymore
                    # We'll let a rollback failure kill the Scheduler (outer shell script should re-start it)
                    self.session.rollback()
            except KeyboardInterrupt:
                self.log.error("KeyboardInterrupt, killing tasks")
                for etl_task_id in self.etl_task_by_id:
                    etl_task = self.etl_task_by_id[etl_task_id]
                    if etl_task.thread_running:
                        self.kill_task(etl_task)
                self.log.error("KeyboardInterrupt done killing tasks")
                raise SystemExit

            if len(self.etl_task_by_id) > 0:
                self.log.debug("Wait {} sec before checking status again".format(self.process_check_interval))
            try:
                time.sleep(self.process_check_interval)
            except KeyboardInterrupt:
                self.log.error("KeyboardInterrupt killing tasks")
                try:
                    self.session.rollback()
                # Different DBs seem to give different exceptions here
                except Exception:  #pylint: disable=broad-except
                    # Already exiting
                    pass
                for etl_task_id in self.etl_task_by_id:
                    etl_task = self.etl_task_by_id[etl_task_id]
                    if etl_task.thread_running:
                        try:
                            self.kill_task(etl_task)
                        except Exception: #pylint: disable=broad-except
                            # Already exiting
                            pass
                self.log.error("KeyboardInterrupt done killing tasks")
                raise SystemExit


if __name__ == '__main__':
    scheduler = Scheduler()
    scheduler.start_monitoring()



"""
Created on May 20, 2015

@author: Derek Wood
"""
import copyreg
import fnmatch
import getpass
import importlib
import inspect
import logging
import pickle as pickle
import pkgutil
import socket
import textwrap
import time
import traceback
import types
from datetime import datetime

# noinspection PyUnresolvedReferences
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.functions import func

from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base
from bi_etl.scheduler.models import (ETL_Tasks,
                                     ETL_Task_Params,
                                     ETL_Task_Status_CD,
                                     ETL_Task_Log,
                                     ETL_Task_Stats,
                                     ETL_Scheduler)
from bi_etl.scheduler.task import ETLTask, Status
from bi_etl.timer import Timer
from bi_etl.utility import log_logging_level, dict_to_str


#################
# Register a way of pickling and unpickling class methods
def _pickle_method(method):
    func_name = method.__func__.__name__
    obj = method.__self__
    cls = method.__self__.__class__
    return _unpickle_method, (func_name, obj, cls)


def _unpickle_method(func_name, obj, cls):
    funct = None
    for cls in cls.mro():
        try:
            funct = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    if funct:
        return funct.__get__(obj, cls)
    else:
        raise RuntimeError('function {} not found'.format(func_name))


copyreg.pickle(types.MethodType, _pickle_method, _unpickle_method)


class SchedulerInterface(object):
    """
    Light scheduler interface that only interacts with the database.
    """
    CLASS_VERSION = 1.0
    CONFIG_SECTION = 'Scheduler'
    scan_etl_classes_performed = False
    SHARED_etl_task_classes = dict()
    SCHEDULER_ETL_JOBS_PACKAGE = 'bi_etl.scheduler.scheduler_etl_jobs'

    def __init__(self,
                 config: BI_ETL_Config_Base,
                 session=None,
                 log=None,
                 scheduler_id=None,
                 allow_create=False
                 ):

        self.config = config
        if log is None:
            self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        else:
            self.log = log
        log_logging_level(self.log)

        self.etl_task_classes = SchedulerInterface.SHARED_etl_task_classes
        base_module_str = self.config.bi_etl.task_finder_base_module
        if base_module_str is None:
            self.base_modules = []
        else:
            if ',' in base_module_str:
                self.base_modules = [s.strip() for s in base_module_str.split(',')]
            else:
                self.base_modules = [base_module_str]
        if SchedulerInterface.SCHEDULER_ETL_JOBS_PACKAGE not in self.base_modules:
            self.base_modules.append(SchedulerInterface.SCHEDULER_ETL_JOBS_PACKAGE)

        ##############
        # Change the schema of the ETL task definition tables
        metadata = ETL_Tasks.__base__.metadata  # @UndefinedVariable pylint: disable=no-member

        if self.config.bi_etl.scheduler is None:
            raise ValueError("Config does not contain bi_etl.scheduler setting")

        # Pickup schema name from config
        schema = self.config.bi_etl.scheduler.db.database_schema
        self.schema = schema
        if self.schema is not None:
            self.log.info("Using etl_tasks tables in schema {}".format(schema))

            # We don't seem to need to change this.  In fact if we added a table dynamically we'd
            # probably want it to initially link in to the other public. tables
            # metadata.schema = 'onr_dw2'

            for tbl_index in metadata.tables:
                tbl = metadata.tables[tbl_index]
                tbl.schema = schema
                tbl.fullname = tbl.fullname.replace('public', schema)

            # Change schema on Sequences
            # pylint: disable=protected-access
            for seq_index in metadata._sequences:
                seq = metadata._sequences[seq_index]
                seq.schema = schema
            #
            # End  Change the schema
            ##############

        if session is None:
            scheduler_db = self.config.bi_etl.scheduler.db
            if scheduler_db is not None:
                self.log.debug("Making new session")
                self.session = scheduler_db.session()
            else:
                self.log.debug("Missing database option in INI file. No session creation possible")
                self.session = None
        else:
            self.log.debug("Using provided session {}".format(repr(session)))
            self.session = session
        del session

        if self.session is None:
            self.scheduler_row = None
            self.log.debug('scheduler_row not retried')
        else:
            if scheduler_id is None:
                scheduler_host_name = self.config.bi_etl.scheduler.host_name
                # Pass null scheduler_host_name to get_scheduler_id_for_host,
                # it looks up in the config or defaults to local host
                self.scheduler_row = self._get_scheduler_row_for_host(
                    qualified_host_name=scheduler_host_name,
                    allow_create=allow_create,
                )
            else:
                self.log.info('Using assigned scheduler ID {}.'.format(scheduler_id))
                self.scheduler_row = self.get_scheduler_row_for_id(scheduler_id)

            if self.scheduler_row is None:
                msg = """
                      Scheduler not found by name (ini Scheduler section : host setting).'
                      Defaulting to scheduler ID 1, if that exists.
                      """
                self.log.warning(textwrap.dedent(msg))
                self.scheduler_row = self.get_scheduler_row_for_id(1)

            self.scheduler_id = self.scheduler_row.scheduler_id
            self.log.info("scheduler_id = {}".format(self.scheduler_id))

    @property
    def ETL_Tasks(self):
        return ETL_Tasks

    @property
    def ETL_Task_Status_CD(self):
        return ETL_Task_Status_CD

    @property
    def ETL_Task_Params(self):
        return ETL_Task_Params

    @property
    def ETL_Task_Log(self):
        return ETL_Task_Log

    @property
    def ETL_Task_Stats(self):
        return ETL_Task_Stats

    @property
    def qualified_host_name(self):
        """
        Gets the qualified host name of the scheduler server.
        """
        return self.scheduler_row.qualified_host_name

    @staticmethod
    def _get_local_qualified_host_name():
        """
        Gets the qualified host name of the current server (not the scheduler server)
        """
        # =======================================================================
        # socket.getfqdn can return different values for each call if the host has multiple aliases.
        # In our case it alternated between values, however it seems like it could be more random than that.
        # So we'll get all aliases pick the lowest value (as a way of trying to be consistent).
        #
        # Note: You can also specify the hostname directly in the config to avoid this.
        # =======================================================================
        try:
            hostname, aliases, ipaddrs = socket.gethostbyaddr(socket.gethostname())
            aliases.append(hostname)

            return min([name for name in aliases if '.' in name])
        except socket.error:
            return socket.gethostname()

    def get_scheduler_row_for_id(self, scheduler_id):
        query = self.session.query(ETL_Scheduler).filter(ETL_Scheduler.scheduler_id == scheduler_id)
        scheduler_row = query.one()
        return scheduler_row

    def _get_scheduler_row_for_host(self, qualified_host_name: str = None, allow_create: bool = False):
        qualified_host_name = (qualified_host_name
                               or
                               self.config.bi_etl.scheduler.qualified_host_name
                               or
                               self._get_local_qualified_host_name()
                               )
        host_name = socket.gethostname() or qualified_host_name
        self.log.debug('get_scheduler_id_for_host: host name is {}'.format(host_name))
        try:
            query = self.session.query(ETL_Scheduler).filter(ETL_Scheduler.qualified_host_name == qualified_host_name)
            scheduler_row = query.one()
            return scheduler_row
        except NoResultFound:
            if allow_create:
                scheduler_row = ETL_Scheduler()
                try:
                    # We make the assumption here that two schedulers won't be initializing for the first
                    # time at the same instant. If they do, they'd get the same ID and one would fail 
                    # on the commit below (PK error)
                    max_row = self.session.query(func.max(ETL_Scheduler.scheduler_id).label("max_id")).one()
                    max_id = max_row.max_id
                except NoResultFound:
                    max_id = 0
                scheduler_row.scheduler_id = (max_id or 0) + 1
                scheduler_row.qualified_host_name = qualified_host_name
                scheduler_row.host_name = host_name
                self.session.add(scheduler_row)
                self.session.commit()
                return scheduler_row
            else:
                return None

    def get_scheduler_id_for_host(self, qualified_host_name=None, allow_create=False):
        scheduler_row = self._get_scheduler_row_for_host(qualified_host_name=qualified_host_name,
                                                         allow_create=allow_create
                                                         )
        if scheduler_row:
            return scheduler_row.scheduler_id
        else:
            return None

    def get_next_task_id(self):
        # We have to query the sequence each time because this code might be running
        # on a different server or thread from others
        next_id = self.session.execute(ETL_Tasks.task_id.default)
        return next_id

    def add_task_by_exact_name(
            self,
            module_name,
            class_name=None,
            display_name=None,
            parent_task_id=None,
            root_task_id=None,
            scheduler_id=None,
            parameters=None,
            submit_by_user_id=None,
            commit=True,
    ):
        """
        Add a task to the scheduler using a module name
        """
        msg = "add_task_by_exact_name module_name={module_name}, class_name={class_name} parent_task_id={parent_task_id} root_task_id={root_task_id}"
        self.log.debug(msg.format(module_name=module_name,
                                  class_name=class_name,
                                  parent_task_id=parent_task_id,
                                  root_task_id=root_task_id,
                                  )
                       )
        task_rec = ETL_Tasks()
        if scheduler_id is None:
            scheduler_id = self.scheduler_id
        task_rec.scheduler_id = scheduler_id
        task_rec.modulename = module_name
        task_rec.classname = class_name
        task_rec.display_name = display_name or module_name
        self.log.debug("--- Getting next task_id")
        task_rec.task_id = task_id = self.get_next_task_id()
        self.log.debug("--- New task_id={}".format(task_id))
        # We don't want self reference parent or root
        if parent_task_id == task_id:
            parent_task_id = None
        if root_task_id == task_id:
            root_task_id = None
        task_rec.parent_task_id = parent_task_id

        parent_task = None
        # Check if we need to fill in the root_task_id by looking up the parent
        if parent_task_id is not None and root_task_id is None:
            parent_task = self.get_task_record(parent_task_id)
            if parent_task is None:
                raise ValueError("--- parent_task_id {} is invalid".format(parent_task_id))
            elif parent_task.root_task_id is not None:
                root_task_id = parent_task.root_task_id
            else:
                root_task_id = parent_task_id
        task_rec.root_task_id = root_task_id

        if submit_by_user_id is not None:
            task_rec.submit_by_user_id = submit_by_user_id
        else:
            if parent_task_id is None:
                task_rec.submit_by_user_id = getpass.getuser()
            else:
                if parent_task is None:
                    parent_task = self.get_task_record(parent_task_id)
                task_rec.submit_by_user_id = parent_task.submit_by_user_id

        task_rec.Status = Status.new
        if parameters is not None:
            self.add_task_paramters(task_rec, parameters, commit=False)
        self.log.debug("--- Adding task to db {}".format(dict_to_str(task_rec)))
        self.session.add(task_rec)
        if commit:
            self.session.commit()
        else:
            self.session.flush()
        return task_id

    def add_task_by_class(
            self,
            etl_task_class_type,
            display_name=None,
            parent_task_id=None,
            root_task_id=None,
            scheduler_id=None,
            parameters=None,
            submit_by_user_id=None,
            commit=True,
    ):
        """
        Add a task to the scheduler using an instance of the task class type.

        Returns
        -------
        task_id: int
        """
        # Test that we got a class type
        if not isinstance(etl_task_class_type, type):
            # Turn name into a class type
            etl_task_class_type = self.find_etl_class_type(etl_task_class_type)
        return self.add_task_by_exact_name(
            module_name=etl_task_class_type.__module__,
            class_name=etl_task_class_type.__name__,
            display_name=display_name,
            parent_task_id=parent_task_id,
            root_task_id=root_task_id,
            scheduler_id=scheduler_id,
            parameters=parameters,
            submit_by_user_id=submit_by_user_id,
            commit=commit,
        )

    def scan_etl_classes(self):
        SchedulerInterface.scan_etl_classes_performed = True
        self.etl_task_classes.clear()
        self.log.debug('scan_etl_classes: base_modules={}'.format(self.base_modules))

        for base_module in self.base_modules:
            base_module_spec = importlib.util.find_spec(base_module)
            self._scan_etl_classes_in_base(base_module_spec)

    def _scan_etl_classes_in_base(self, base_module_spec):
        path = base_module_spec.submodule_search_locations
        prefix = base_module_spec.name + '.'
        self.log.debug('_scan_etl_classes_in_base: scanning path={}'.format(path))
        self.log.debug('_scan_etl_classes_in_base: using prefix={}'.format(prefix))
        for (module_finder, name, ispkg) in pkgutil.walk_packages(path=path,
                                                                  prefix=prefix,
                                                                  onerror=lambda pkg: self.log.debug(
                                                                      'Error importing package {}'.format(pkg))
                                                                  ):
            if not ispkg:
                module = None
                try:
                    self.log.debug('_scan_etl_classes_in_base: Loading module {}'.format(name))
                    module_loader = module_finder.find_spec(name).loader
                    module = module_loader.load_module(name)
                except (Exception, SystemExit) as e:  # pylint: disable=broad-except
                    self.log.debug('_scan_etl_classes_in_base: Skipping {} due to {}'.format(name, e))
                if module is not None:
                    try:
                        self.log.debug('_scan_etl_classes_in_base: Checking for matching class(es) in {}'.format(name))
                        task_class_dict = self._find_etl_classes_in_module(module)
                        if task_class_dict is not None:
                            self.etl_task_classes.update(task_class_dict)
                    except ValueError as e:
                        self.log.debug('_scan_etl_classes_in_base: Skipping {} due to {}'.format(name, e))

    def _find_etl_classes_in_module(self, module):
        """
        Returns a dictionary of the ETLTask instances defined in a given module
        """
        class_matches_by_type = dict()

        # Look for ETLTask inherited classes
        for _, classObj in inspect.getmembers(module, inspect.isclass):
            # Check that the class is defined in our module and not imported
            if classObj.__module__ == module.__name__:
                baseclasses = inspect.getmro(classObj)
                if ETLTask in baseclasses and str(classObj) != str(ETLTask):
                    inst = classObj()
                    qualified_name = inst.name
                    self.log.info('_find_etl_classes_in_module found {}'.format(qualified_name))
                    class_matches_by_type[qualified_name] = classObj

        return class_matches_by_type

    def _find_etl_class_in_module(self, module, class_name=None):
        class_matches_by_type = self._find_etl_classes_in_module(module)
        task_class = None
        if len(class_matches_by_type) == 1:
            name, task_class = class_matches_by_type.popitem()
            self.log.debug("_find_etl_class_in_module found 1 ETLTask matching class {}".format(name))
        else:
            if class_name is None:
                class_name = module.__name__.split('.')[-1]
            self.log.debug("_find_etl_class_in_module found ETLTask classes {}".format(class_matches_by_type))
            for name in class_matches_by_type:
                if name.lower() == class_name.lower():
                    task_class = class_matches_by_type[name]
                    self.log.debug("_find_etl_class_in_module found class by name {}={}".format(task_class, class_name))
        if task_class is None:
            msg = "Module {} doesn't contain a single ETLTask class, could not choose class from {}".format(module,
                                                                                                            class_matches_by_type)
            self.log.debug(msg)
            raise ValueError(msg)
        return task_class

    def _get_class_from_qualified_name(self, qualified_name):
        parts = qualified_name.split('.')
        module_part = '.'.join(parts[:-1])
        class_part = parts[-1:][0]
        module = importlib.import_module(module_part)
        etl_class = self._find_etl_class_in_module(module, class_part)
        return etl_class

    def find_etl_classes(self, partial_module_name):
        if not SchedulerInterface.scan_etl_classes_performed:
            # Since we don't have a class database, try and quickly import the module assuming it's fully qualified
            # TODO: Once classes and dependency hierarchy is stored in the database, scan that instead.
            try:
                # We try to get the class simply to verify it exists
                _ = self._get_class_from_qualified_name(partial_module_name)
                return [partial_module_name]  # Caller wants the qualified name not the instance
            except (ImportError, ValueError):
                pass
            self.scan_etl_classes()

        if not isinstance(partial_module_name, str):
            partial_module_name = str(partial_module_name)

        matches = list()
        for etl_class in self.etl_task_classes:
            if (fnmatch.fnmatch(etl_class, partial_module_name)
                    or fnmatch.fnmatch(etl_class, '*.' + partial_module_name + '.*')
                    or fnmatch.fnmatch(etl_class, '*.' + partial_module_name)
                    or fnmatch.fnmatch(etl_class, partial_module_name + '.*')
            ):
                matches.append(etl_class)
        return matches

    def find_etl_class_name(self, partial_module_name):
        matches = self.find_etl_classes(partial_module_name)
        if len(matches) == 1:
            return matches[0]
        else:
            raise ValueError(
                'partial_module_name {} matched {} and not a single record'.format(partial_module_name, matches))

    def get_etl_class_instance(self, qualified_name):
        if not SchedulerInterface.scan_etl_classes_performed:
            # Since we don't have a class database, try and quickly import the module assuming it's fully qualified
            etl_class = self._get_class_from_qualified_name(qualified_name)
            return etl_class
        return self.etl_task_classes[qualified_name]

    def find_etl_class_instance(self, partial_module_name):
        class_name = self.find_etl_class_name(partial_module_name)
        return self.get_etl_class_instance(class_name)

    def find_etl_class_type(self, partial_module_name):
        class_name = self.find_etl_class_name(partial_module_name)
        return self.get_etl_class_instance(class_name).__class__

    def add_task_by_partial_name(self,
                                 partial_module_name,
                                 display_name=None,
                                 parent_task_id=None,
                                 root_task_id=None,
                                 scheduler_id=None,
                                 parameters=None,
                                 submit_by_user_id=None,
                                 commit=True,
                                 ):
        """
        Add a task to the scheduler using the module_name (partial) and optionally class_name

        Returns
        -------
        task_id: int
        """
        etl_task_class_name = self.find_etl_class_name(partial_module_name)
        return self.add_task_by_exact_name(etl_task_class_name,
                                           display_name=display_name,
                                           parent_task_id=parent_task_id,
                                           root_task_id=root_task_id,
                                           scheduler_id=scheduler_id,
                                           parameters=parameters,
                                           submit_by_user_id=submit_by_user_id,
                                           commit=commit,
                                           )

    def get_task_record(self, task_id):
        # Make sure task_id is an integer
        task_id = int(task_id)
        return self.session.query(ETL_Tasks).get(task_id)

    def get_task_status(self, task_id):
        """
        Gets the Status of a task

        Returns
        -------
        status_of_task: bi_etl.scheduler.status.Status
        """
        return self.get_task_record(task_id).Status

    def wait_for_task(self, task_id, check_interval=1, max_wait=None):
        """
        Waits for a task to finish.

        Parameters
        ----------
        task_id: int
            The task_id to wait for

        check_interval: float
            How many seconds to wait between checks.
            The argument may be a floating point number for subsecond precision.
            
        max_wait: float
            The maximum seconds to wait for the job to finish.  If None or 0, will wait indefinitely.

        Returns
        -------
        status_of_task: bi_etl.scheduler.status.Status
        """
        timer = Timer()
        while not self.get_task_status(task_id).is_finished():
            if max_wait:
                if timer.seconds_elapsed > max_wait:
                    raise TimeoutError(
                        'wait_for_task wait time {} exceeded max_wait {}'.format(timer.seconds_elapsed, max_wait))
            time.sleep(check_interval)
        return self.get_task_status(task_id)

    def add_task_paramter(self, task, parameter_name, parameter_value, commit=True):
        if isinstance(task, int):
            task_rec = self.get_task_record(task)
        elif isinstance(task, ETL_Tasks):
            task_rec = task
        else:
            raise ValueError("add_task_paramter task parameter must be int or ETL_Tasks")
        task_rec.parameters.append(ETL_Task_Params(parameter_name, pickle.dumps(parameter_value)))
        if commit:
            self.session.commit()

    def add_task_paramters(self, task, parameters, commit=True):
        if isinstance(task, int):
            task_rec = self.get_task_record(task)
        elif isinstance(task, ETL_Tasks):
            task_rec = task
        else:
            raise ValueError("add_task_paramters task parameter must be int or ETL_Tasks")
        if isinstance(parameters, list):
            for parameter_name, parameter_value in parameters:
                self.add_task_paramter(task_rec, parameter_name, parameter_value, commit=commit)
        else:  # assume it's a dict
            for parameter_name in parameters:
                parameter_value = parameters[parameter_name]
                self.add_task_paramter(task_rec, parameter_name, parameter_value, commit=commit)
        if commit:
            self.session.commit()

    def get_task_parameter_dict(self, task) -> dict:
        """

        Parameters
        ----------
        task: int or ETL_Tasks

        Returns
        -------
        dict
        """
        if isinstance(task, int):
            task_id = task
            task_rec = self.get_task_record(task_id)
        elif isinstance(task, ETL_Tasks):
            task_rec = task
            task_id = task_rec.task_id
            assert task_id is not None, "get_task_parameter_dict called for task with no task_id {}".format(task_rec)
        else:
            raise ValueError('Unexpected type for task parameter {}'.format(type(task)))
        self.log.debug("Getting parameters for task id {}".format(task_id))
        params_dict = dict()
        for param_rec in task_rec.parameters:
            try:
                param_value = pickle.loads(param_rec.param_value)
                params_dict[param_rec.param_name] = param_value
            except pickle.UnpicklingError:
                traceback.print_exc()
                if len(param_rec.param_value) > 200:
                    msg = "ETL Task {} unable to unpickle {} value (too long to print. len={})"
                    msg = msg.format(repr(task_id),
                                     param_rec.param_name,
                                     len(param_rec.param_value)
                                     )
                    self.log.error(msg)
                else:
                    msg = "ETL Task {} unable to unpickle {} value '{}'"
                    msg = msg.format(repr(task_id),
                                     param_rec.param_name,
                                     param_rec.param_value
                                     )
                    self.log.error(msg)
                raise
        return params_dict

    def load_parameters(self, etl_task):
        """
        This is done in the etl_task thread/process (called by run method)
        so that the parameters don't have to be loaded into the scheduler thread or passed between threads/processes.
        """
        params_dict = self.get_task_parameter_dict(etl_task.task_id)
        for param_name in params_dict:
            etl_task.set_parameter(param_name, params_dict[param_name], local_only=True)

    def get_jobs_by_root_id(self, root_task_id):
        query = self.session.query(ETL_Tasks).filter(ETL_Tasks.root_task_id == root_task_id)
        return query

    def get_status_code_list(self):
        return self.session.query(ETL_Task_Status_CD).all()

    def get_heartbeat_time(self):
        if self.scheduler_row is not None:
            return self.scheduler_row.last_heartbeat
        else:
            return None

    def get_heartbeat_age_timedelta(self):
        last_heartbeat = self.get_heartbeat_time()
        if last_heartbeat is None:
            return None
        else:
            return (datetime.now() - last_heartbeat)

    def heatbeat_now(self):
        self.scheduler_row.last_heartbeat = datetime.now()
        self.session.flush()

    def __reduce__(self):
        return (SchedulerInterface, ())



from bi_etl.notifiers.notifier_base import NotifierBase


class LogNotifier(NotifierBase):
    def __init__(self):
        super().__init__()

    def send(self, subject, message, sensitive_message=None, attachment=None, throw_exception=False):
        if subject is not None:
            self.log.info(subject)
        if message is not None:
            self.log.info(message)



import logging
import random
import subprocess
import time
from pathlib import Path
from typing import Optional, Union

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_types.path_types import ExecutablePath

log = logging.getLogger('etl.utils.ssh_forward')


def ssh_run_command(
        host: str,
        user: str,
        command: str,
        ssh_path: Optional[Union[str, Path]] = None,
        ) -> str:
    # Command line options documentation
    # https://man.openbsd.org/ssh
    if ssh_path is None:
        ssh_path = 'ssh'
    cmd = [
        ssh_path,
        '{user}@{host}'.format(user=user, host=host),
        command,
    ]
    log.debug("Starting ssh")
    log.debug(' '.join(cmd))
    try:
        stdout = subprocess.PIPE
        stderr = subprocess.PIPE
        log.debug(f"Running {' '.join(cmd)}")
        p = subprocess.Popen(cmd, stdout=stdout, stderr=stderr, universal_newlines=True)
        log.info("Started ssh as ppid {}".format(p.pid))
        outs, errs = p.communicate()
        rc = p.poll()
        if rc is not None:
            if rc != 0:
                if outs:
                    log.info(outs)
                if errs:
                    log.error(errs)
                log.error("ssh return code = {}".format(rc))
        return outs

    except subprocess.CalledProcessError as e:
        log.error(e.stdout)
        log.error(e.stderr)
        raise e


class SSH_Config(ConfigHierarchy):
    host: str
    user_id: str
    ssh_path: ExecutablePath = None


def ssh_run_command_using_config(config: SSH_Config, command: str) -> str:
    return ssh_run_command(
        host=config.host,
        user=config.user_id,
        ssh_path=config.ssh_path,
        command=command,
    )


def ssh_forward(
        host: str,
        user: str,
        server: str,
        server_port: int,
        local_port: int = None,
        wait: bool = False,
        ssh_path: Optional[Union[str, Path]] = None,
        seconds_wait_for_usage: int = 60):
    # Command line options documentation
    # https://man.openbsd.org/ssh
    if ssh_path is None:
        ssh_path = 'ssh'
    if local_port is None:
        local_port = random.randrange(10000, 60000)
    cmd = [
        ssh_path,
        '{user}@{host}'.format(user=user, host=host),
        '-L', '127.0.0.1:{local_port}:{server}:{server_port}'.format(
            local_port=local_port,
            server=server,
            server_port=server_port,
        ),
        '-o', 'ExitOnForwardFailure=yes',
        '-o', 'StrictHostKeyChecking=no',
        # Sleep 10 will give 10 seconds to connect before it shuts down.
        # It will not exit while the forward port is in use.
        # If this is the second forward to run it will exit after 10 seconds since it won't have
        # an active forward.
        'sleep', str(seconds_wait_for_usage),
    ]
    log.info(f'Starting ssh to {host}:{server_port}')
    log.debug(' '.join(cmd))
    try:
        if wait:
            stdout = subprocess.PIPE
            stderr = subprocess.PIPE
        else:
            stdout = subprocess.DEVNULL
            stderr = subprocess.DEVNULL
        p = subprocess.Popen(cmd, stdout=stdout, stderr=stderr, universal_newlines=True)
        log.info("Started ssh as ppid {}".format(p.pid))
        if wait:
            outs, errs = p.communicate()
        else:
            time.sleep(0.25)
            outs = None
            errs = None
        rc = p.poll()
        if rc is not None:
            if rc != 0:
                if outs:
                    log.info(outs)
                if errs:
                    log.error(errs)
                log.error("ssh return code = {}".format(rc))
        else:
            log.info("ssh tunnel running OK with local_port = {}".format(local_port))

        return local_port

    except subprocess.CalledProcessError as e:
        log.error(e.stdout)
        log.error(e.stderr)
        raise e


class SSH_Forward_Config(SSH_Config):
    server: str
    server_port: int
    local_port: int
    seconds_wait_for_usage: int = 60
    seconds_wait_for_tunnel_start: int = 2


def ssh_forward_using_config(config: SSH_Forward_Config, wait=False):
    local_port = ssh_forward(
        host=config.host,
        user=config.user_id,
        local_port=config.local_port,
        server=config.server,
        server_port=config.server_port,
        ssh_path=config.ssh_path,
        seconds_wait_for_usage=config.seconds_wait_for_usage,
        wait=wait,
    )
    if config.seconds_wait_for_tunnel_start:
        time.sleep(config.seconds_wait_for_tunnel_start)

    return local_port



import logging
import os
import subprocess
import tempfile
from pprint import pformat

log = logging.getLogger('etl.utils.psql_command')


def generate_extract_sql(table_name, sql_file_path, output_file_path, delimiter='\\013', null='', encoding='UTF-8'):
    with open(sql_file_path, "wt") as f:
        f.write("\copy {table_name} to '{output_file_path}' with delimiter E'{delimiter}' NULL '{null}' ENCODING '{encoding}';".format(
            table_name=table_name,
            output_file_path=output_file_path,
            delimiter=delimiter,
            null=null,
            encoding=encoding,
        ))

    """  Commands to re-encode content as UTF-16
    powershell -c "Get-Content -Encoding utf8  -TotalCount 1 .\output.txt | Set-Content -Encoding Unicode output-utf16le.txt
    powershell -c "Get-Content -Encoding utf8 .\output.txt | Add-Content -Encoding Unicode output-utf16le.txt
    """


def psql(config, dbname, username, password, sql_file_path):
    cmd = [config.get('psql', 'path', fallback='psql'),
           '--dbname', dbname,
           '--username', username,
           # Password must be set in %APPDATA%\postgresql\pgpass.conf
           # For example C:\Users\__developer__\AppData\Roaming\postgresql\pgpass.conf
           '--file', sql_file_path,
           ]
    log.debug([x for x in cmd])
    messages = list()

    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        # while p.poll() is None:
        #     # TODO: Use pexpect instead?  bcp output to the PIPE is buffered when using this method.
        #     line = p.stdout.readline()
        #     if line != b'':
        #         line = line.strip()
        #         messages.append(line)
        #         self.log.info(line)
        # self.log.info('--psql exit--')
        # for line in p.stdout.readlines():
        #     if line != b'':
        #         line = line.strip()
        #         messages.append(line)
        #         self.log.info(line)
        outs, errs = p.communicate(input=password + "\n")
        # outs, errs = p.communicate()
        rc = p.returncode
        if rc != 0:
            log.error('psql returned code {}'.format(rc))
            log.error('psql returned message {}'.format(outs))
            log.error('psql returned message {}'.format(errs))
            raise RuntimeError('psql execution error')
        else:
            log.debug('psql returned message {}'.format(outs))
            log.debug('psql returned error {}'.format(errs))
            if outs.startswith(b'COPY '):
                rows = int(outs[5:])
                log.debug("Rows message found rows = {}".format(rows))
                return rows
            else:
                log.error('psql returned message {}'.format(outs))
                log.error('psql returned error {}'.format(errs))
                raise ValueError("psql output did not contain COPY rows")
    except subprocess.CalledProcessError as e:
        log.error("Error code " + str(e.returncode))
        log.error('BCP messages:')
        log.error(pformat(messages))
        log.error('BCP output:')
        log.error(e.output)
        log.error('-' * 80)
        raise RuntimeError("psql error code " + str(e.returncode))


def psql_extract(self, dbname, username, password, table_name, output_file_path, delimiter='\\013', null='', encoding='UTF-8'):
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
        sql_file_path = os.path.join(temp_dir, 'extract.sql')
        self.generate_extract_sql(table_name, sql_file_path, output_file_path, delimiter, null, encoding)
        self.psql(dbname, username, password, sql_file_path)


"""
Created on May 15, 2015

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import dbm
import math
import os
import pickle
import shelve
import string
import sys
import tempfile
import typing
import weakref

import semidbm

from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base
from bi_etl.lookups.lookup import Lookup
from bi_etl.memory_size import get_dir_size
from bi_etl.memory_size import get_size_gc

if typing.TYPE_CHECKING:
    from bi_etl.components.etlcomponent import ETLComponent


__all__ = ['DiskLookup']


class DiskLookup(Lookup):
    DEFAULT_PATH = None
    
    def __init__(self,
                 lookup_name: str,
                 lookup_keys: list,
                 parent_component: ETLComponent,
                 config: BI_ETL_Config_Base = None,
                 use_value_cache: bool = True,
                 path=None,
                 init_parent: bool = True,
                 **kwargs):
        """
        Optional parameter path where the lookup files should be persisted to disk
        """
        if init_parent:
            super().__init__(
                lookup_name=lookup_name,
                lookup_keys=lookup_keys,
                parent_component=parent_component,
                use_value_cache=use_value_cache,
                config=config,
                **kwargs
                )
        self._set_path(path)
        self.dbm = None
        self._cache_dir_mgr = None
        self._cache_file_path = None
        self.use_value_cache = False
        self._finalizer = None
        # Shelf requires str keys
        self._hashable_key_type = str
        
    def _set_path(self, path: str):
        if path is not None:
            self.path = path
        else:
            if self.config is not None:
                self.path = self.config.get('Cache', 'path', fallback=DiskLookup.DEFAULT_PATH)
            else:
                self.path = DiskLookup.DEFAULT_PATH
            
    def init_cache(self):
        if self.cache_enabled is None:
            self.cache_enabled = True
        if self.cache_enabled:
            file_prefix = ''.join([c for c in self.lookup_name if c in string.ascii_letters])
            self._cache_dir_mgr = tempfile.TemporaryDirectory(
                dir=self.path,
                prefix=file_prefix,
                ignore_cleanup_errors=True,
            )
            self._cache_file_path = self._cache_dir_mgr.name
            self.log.info("Creating cache in {}".format(self._cache_file_path))
            if sys.platform.startswith('win'):
                self.dbm = semidbm.open(self._cache_file_path, 'n')
            else:
                file = os.path.join(self._cache_file_path, 'data')
                self.dbm = dbm.open(file, 'n')
            self._cache = shelve.BsdDbShelf(
                self.dbm,
                protocol=pickle.HIGHEST_PROTOCOL,
                writeback=False,
            )
            self._finalizer = weakref.finalize(self, self._cleanup)

    def __len__(self):
        if self._cache is not None:
            return len(self.dbm.keys())
        else:
            return 0
        
    def _get_first_row_size(self, row: Lookup.ROW_TYPES):
        if self.dbm:
            # Slow but shouldn't be too bad twice
            self._row_size = get_size_gc(self.dbm)
        
    def check_estimate_row_size(self, force_now: bool = False):
        if force_now or not self._done_get_estimate_row_size:
            row_cnt = min(len(self), 1000)
            total_row_sizes = 0
            row_num = 0
            for row in self:
                total_row_sizes += get_size_gc(self.get_hashable_combined_key(row))
                row_num += 1
                if row_num >= row_cnt:
                    break      
            self._row_size = math.ceil(total_row_sizes / row_cnt)
            msg = '{lookup_name} row key size (in memory) now estimated at {size:,} bytes per row'
            msg = msg.format(lookup_name=self.lookup_name,
                             size=self._row_size)
            self.log.debug(msg)
            self._done_get_estimate_row_size = True   
        
    def get_disk_size(self):
        if self._cache_file_path:
            return get_dir_size(self._cache_file_path)
        else:
            return 0

    def _cleanup(self):
        if self._cache is not None:
            self.log.debug(f"Cleanup cache file {self._cache_dir_mgr} {self._cache_dir_mgr.name}")
        self.clear_cache()            

    def clear_cache(self):
        if self._cache is not None:
            self._cache.close()
            self._cache_dir_mgr.cleanup()
        self._cache = None



"""
Created on Oct 9, 2015

@author: Derek Wood
"""
import io
import typing
from collections import defaultdict
from decimal import Context, ROUND_HALF_EVEN
from operator import itemgetter
from sys import stdout

from bi_etl.components.etlcomponent import ETLComponent
from bi_etl.conversions import str2decimal, str2date
from bi_etl.scheduler.task import ETLTask
from bi_etl.utility import get_integer_places


# noinspection PyBroadException
class DataAnalyzer(ETLComponent):
    """
    Class that analyzes the data rows passed to it. 
    * Tracks distinct columns passed in
    * Tracks datatype of each column
    * Tracks valid values of each column
    
    Parameters
    ----------
    task: ETLTask
        The  instance to register in (if not None)
    logical_name: str
        The logical name of this source. Used for log messages.
    """
    COLUMN_HEADERS_DICT = {
        'col': 'Column Name',
        'type': 'Data Type',
        'non_null_rows': 'Non-Null Rows',
        'cardinality': 'Cardinality',
        'msg': 'Message',
        'present': 'Rows with this column',
        'not_present_on_rows': 'Rows without this column',
        'most_common_value': 'Most Common Value',
    }
    COLUMN_HEADERS_FORMAT = "{col:45}  {type:20} {non_null_rows:>15} {cardinality:>15}   {msg}"
    COLUMN_HEADERS = COLUMN_HEADERS_FORMAT.format(**COLUMN_HEADERS_DICT)
    DEFAULT_FORMAT = "{col:45}  {type:20} {non_null_rows:15,} {cardinality:15,}   {msg}"
    EQUALS_FORMAT = "{col:45} type = {type:20} non_null_rows={non_null_rows:15,} cardinality={cardinality:15,} {msg}"
    PIPE_FORMAT = "{col}|{type}|{present}|{not_present_on_rows}|{non_null_rows}|{cardinality}|{most_common_value}|{msg}"
    PIPE_HEADERS = PIPE_FORMAT.format(**COLUMN_HEADERS_DICT)

    class DataType(object):
        def __init__(self, name, length=None, precision=None, fmt=None):
            self.name = name
            self.length = length
            self.precision = precision
            self.format = fmt

        def __repr__(self):
            return "{}({},{},fmt={})".format(self.name, self.length, self.precision, self.format)

        def __str__(self):
            if self.length is None and self.format is None:
                return self.name
            if self.format is not None:
                return "{}({})".format(self.name, self.format)
            elif self.precision is None:
                return "{}({})".format(self.name, self.length)
            else:
                return "{}({},{})".format(self.name, self.length, self.precision)

    def __init__(self,
                 task: typing.Optional[ETLTask] = None,
                 logical_name: str = 'DataAnalyzer',
                 **kwargs
                 ):
        # Don't pass kwargs up. They should be set here at the end
        super(DataAnalyzer, self).__init__(task=task, logical_name=logical_name)

        self.float_as_decimal = False

        self.rows_processed = 0
        self.column_names = list()
        self.column_valid_values = dict()
        self.column_data_types = dict()
        self.column_data_types_counts = dict()
        self.column_names_consistent = True
        self.new_columns_after_first_row = False
        self.column_present_count = dict()
        self.column_not_null = dict()
        self.duplicate_column_names = dict()
        # Row level storage
        self.row_column_name_set = set()

        # Should be the last call of every init
        self.set_kwattrs(**kwargs)

    def _reset_storage(self):
        self.rows_processed = 0
        self.column_names = list()
        self.column_valid_values = dict()
        self.column_data_types = dict()
        self.column_data_types_counts = dict()
        self.column_names_consistent = True
        self.new_columns_after_first_row = False
        self.column_present_count = dict()
        self.column_not_null = dict()
        self.duplicate_column_names = dict()
        # Row level storage
        self.row_column_name_set = set()

    def __iter__(self):
        return None

    def close(self, error: bool = False):
        if not self.is_closed:
            super(DataAnalyzer, self).close(error=error)
            self._reset_storage()

    def _type_from_value(self, value):
        if isinstance(value, str):
            # Look for numbers in text
            try:
                dec = str2decimal(value)
                # If the value has no fractional digits, return integer.
                # Note: We could use _isinteger() however that calls 1.0 an integer.
                # Whereas a file with 1.0 values indicates possible fractional values
                # Decimal('1.0').as_tuple().exponent returns -1
                # or 
                # Decimal('1.0')._exp returns -1
                (_, digits, exponent) = dec.as_tuple()
                if exponent >= 0:
                    return DataAnalyzer.DataType(name='Integer',
                                                 length=len(digits) + exponent,
                                                 )
                else:
                    return DataAnalyzer.DataType(name='Decimal',
                                                 length=max(len(digits),
                                                            abs(exponent)),
                                                 precision=abs(exponent),
                                                 )
            except Exception:
                pass

            # Look for date in text
            if '-' in value:
                if ':' in value:
                    for dt_format in [
                                      "%Y-%m-%d %H:%M", "%m-%d-%Y %H:%M", "%d-%m-%Y %H:%M",
                                      "%Y-%m-%d %H:%M:%S", "%m-%d-%Y %H:%M:%S", "%d-%m-%Y %H:%M:%S",
                                      "%Y-%m-%d %H:%M:%S.%f", "%m-%d-%Y %H:%M:%S.%f", "%d-%m-%Y %H:%M:%S.%f",
                                      "%Y-%m-%d %I:%M %p", "%m-%d-%Y %I:%M %p", "%d-%m-%Y %I:%M %p",
                                      "%Y-%m-%d %I:%M:%S %p", "%m-%d-%Y %I:%M:%S %p", "%d-%m-%Y %I:%M:%S %p",
                                      ]:
                    # noinspection PyBroadException
                        try:
                            _ = str2date(value, dt_format=dt_format)
                            dt_type = DataAnalyzer.DataType(name="Date")
                            dt_type.format = dt_format
                            dt_type.length = len(value)
                            return dt_type
                        except Exception:
                            pass
                else:  # No time in value
                    for dt_format in ["%Y-%m-%d", "%m-%d-%Y", "%d-%m-%Y",]:
                        try:
                            _ = str2date(value, dt_format=dt_format)
                            dt_type = DataAnalyzer.DataType(name="Date")
                            dt_type.format = dt_format
                            dt_type.length = len(value)
                            return dt_type
                        except Exception:
                            pass
            elif '/' in value:
                if ':' in value:
                    for dt_format in [
                                      "%Y/%m/%d %H:%M", "%m/%d/%Y %H:%M", "%d/%m/%Y %H:%M",
                                      "%Y/%m/%d %H:%M:%S", "%m/%d/%Y %H:%M:%S", "%d/%m/%Y %H:%M:%S",
                                      "%Y/%m/%d %H:%M:%S.%f", "%m/%d/%Y %H:%M:%S.%f", "%d/%m/%Y %H:%M:%S.%f",
                                      "%Y/%m/%d %I:%M %p", "%m/%d/%Y %I:%M %p", "%d/%m/%Y %I:%M %p",
                                      "%Y/%m/%d %I:%M:%S %p", "%m/%d/%Y %I:%M:%S %p", "%d/%m/%Y %I:%M:%S %p",
                                      ]:
                        try:
                            _ = str2date(value, dt_format=dt_format)
                            dt_type = DataAnalyzer.DataType(name="Date")
                            dt_type.format = dt_format
                            dt_type.length = len(value)
                            return dt_type
                        except Exception:
                            pass
                            # Else it's an actual string
                else:  # No time in value
                    for dt_format in ["%Y/%m/%d", "%m/%d/%Y", "%d/%m/%Y"]:
                        try:
                            _ = str2date(value, dt_format=dt_format)
                            dt_type = DataAnalyzer.DataType(name="Date")
                            dt_type.format = dt_format
                            dt_type.length = len(value)
                            return dt_type
                        except Exception:
                            pass
                            # Else it's an actual string
            return DataAnalyzer.DataType(name=type(value).__name__, length=len(value))
        elif isinstance(value, int):
            return DataAnalyzer.DataType(name='Integer', length=get_integer_places(value))
        elif isinstance(value, float):
            if self.float_as_decimal:
                dec = Context(prec=16, rounding=ROUND_HALF_EVEN).create_decimal_from_float(value).normalize()
                (_, digits, exponent) = dec.as_tuple()
                if exponent >= 0:
                    return DataAnalyzer.DataType(name='Integer',
                                                 length=len(digits) + exponent,
                                                 )
                else:
                    return DataAnalyzer.DataType(name='Decimal',
                                                 length=max(len(digits),
                                                            abs(exponent)
                                                            ),
                                                 precision=abs(exponent))

        return DataAnalyzer.DataType(name=type(value).__name__)

    def next_row(self):
        self.row_column_name_set = set()
        self.rows_processed += 1

    def analyze_column(self, column_name, column_value, column_number=None):
        self.column_present_count[column_name] = self.column_present_count.get(column_name, 0) + 1

        # Process column names
        if column_number is not None:
            if len(self.column_names) < column_number:
                self.column_names.append(column_name)
            else:
                if self.column_names[column_number - 1] != column_name:
                    self.column_names_consistent = False
        else:
            if column_name not in self.column_names:
                self.column_names.append(column_name)

        if column_name not in self.row_column_name_set:
            self.row_column_name_set.add(column_name)
        else:
            self.duplicate_column_names.get(column_name, set()).add(column_number)

        # Process column_valid_values
        if column_name not in self.column_valid_values:
            self.column_valid_values[column_name] = dict()
            if self.rows_processed > 0:
                self.new_columns_after_first_row = True
        value = column_value
        try:
            hash(column_value)
        except TypeError:  # unhashable type
            value = str(column_value)

        self.column_valid_values[column_name][value] = self.column_valid_values[column_name].get(value, 0) + 1

        # Process column_data_types
        if column_value is not None:
            self.column_not_null[column_name] = self.column_not_null.get(column_name, 0) + 1

            existing_type = self.column_data_types.get(column_name)
            row_type = self._type_from_value(column_value)

            if column_name not in self.column_data_types_counts:
                self.column_data_types_counts[column_name] = defaultdict(int)
            self.column_data_types_counts[column_name][row_type.name] += 1

            new_type = existing_type
            if existing_type is None:
                new_type = row_type
            else:
                if existing_type.name in ['str', 'unicode', 'bytes']:
                    new_type.length = max(row_type.length, existing_type.length)
                elif existing_type.name == 'Date':
                    if row_type.name == 'Date':
                        new_type.length = max(row_type.length, existing_type.length)
                        if isinstance(existing_type.format, dict):
                            # Add one to the counter for this format
                            new_type.format[row_type.format] = new_type.format.get(row_type.format, 0) + 1
                        elif row_type.format != existing_type.format:
                            fmts = dict()
                            fmts[existing_type.format] = self.rows_processed - 1
                            fmts[row_type.format] = 1
                            new_type.format = fmts
                    elif row_type.name in ['str', 'unicode', 'bytes']:
                        new_type = row_type
                        new_type.length = max(row_type.length, existing_type.length)
                elif existing_type.name == 'Integer':
                    if row_type.name == 'Integer':
                        new_type.length = max(row_type.length, existing_type.length)
                    elif row_type.name == 'Decimal':
                        new_type = row_type
                        new_type.length = max(row_type.length, existing_type.length)
                    else:
                        new_type = DataAnalyzer.DataType(name='str')
                        new_type.length = max(row_type.length, existing_type.length)
                else:
                    if row_type.name != existing_type.name:
                        new_type = DataAnalyzer.DataType(name='str')
                        new_type.length = max(row_type.length, existing_type.length)
            self.column_data_types[column_name] = new_type

    def analyze_row(self, row):
        """
        Analyze the data row passed in. Call this for all the rows that should be analyzed.
        """
        stats = self.get_stats_entry(stats_id='analyze_row')
        stats.timer.start()

        stats['rows processed'] = self.rows_processed
        column_number = 0
        for column_name in row.columns_in_order:
            column_number += 1
            column_value = row[column_name]
            self.analyze_column(column_name=column_name,
                                column_value=column_value,
                                column_number=column_number,
                                )
        self.next_row()
        stats.timer.stop()

    def print_analysis(self,
                       out: io.TextIOBase = None,
                       valid_value_limit: int = 10,
                       columns_header: str = None,
                       columns_out_fmt: str = None
                       ):
        """
        Print the data analysis results.

        Parameters
        ----------
        out:
            The File to write the results to. Default=``stdout``
            valid_value_limit (int): How many valid values should be printed.
        valid_value_limit:
            The number of valid values to output
        columns_header:
            The table header for the columns list
        columns_out_fmt:
            The format to use for lines
        """
        if out is None:
            out = stdout

        if columns_out_fmt is None:
            columns_out_fmt = self.DEFAULT_FORMAT

        if columns_header is None:
            columns_header = self.COLUMN_HEADERS

        print(f"\nRows processed = {self.rows_processed:,}", file=out)
        if not self.column_names_consistent:
            print("**** COLUMN NAMES NOT CONSISTENT IN ALL ROWS", file=out)
        if self.new_columns_after_first_row:
            print("**** NEW COLUMN NAME APPEARED AFTER FIRST ROW", file=out)

        print("Columns:", file=out)
        print(columns_header, file=out)
        column_dict = dict()
        col_cnt = 0
        for col in self.column_names:
            col_cnt += 1

            if col not in column_dict:
                column_dict[col] = list()
            column_dict[col].append(col_cnt)

            msg = ""
            not_present_on_rows = self.rows_processed - self.column_present_count.get(col, 0)
            if not_present_on_rows > 0:
                msg += f" [Not present on {not_present_on_rows:,} rows]"

            most_common_value = None
            vv = self.column_valid_values.get(col)
            if vv is not None:
                vv_list = sorted(list(vv.items()), key=itemgetter(1), reverse=True)
                if len(vv_list) >= 1:
                    most_common_value = vv_list[0]

            print(columns_out_fmt.format(
                col=col,
                type=str(self.column_data_types.get(col)),
                present=self.column_present_count.get(col, 0),
                not_present_on_rows=self.rows_processed - self.column_present_count.get(col, 0),
                non_null_rows=self.column_not_null.get(col, 0),
                cardinality=len(self.column_valid_values.get(col, list())),
                most_common_value=most_common_value,
                msg=msg,
            ),
                file=out
            )

        print("", file=out)
        if len(self.duplicate_column_names) > 0:
            print("Duplicate column names:", file=out)
            for col_name, col_positions in self.duplicate_column_names.items():
                print(f"Column {col_name} appears in positions {col_positions}", file=out)

        print("", file=out)
        print("Columns Valid Values:", file=out)
        col_cnt = 0
        for col in self.column_names:
            col_cnt += 1
            if col_cnt > 1:
                print("", file=out)
            vv = self.column_valid_values.get(col)
            if vv is not None:
                vv_list = sorted(list(vv.items()), key=itemgetter(1), reverse=True)
                print(f"{col} (col {col_cnt}) Cardinality {len(vv):,} Apparent type {str(self.column_data_types.get(col))} Source Type {type(vv_list[0][0]).__name__} :", file=out)
                for (v, freq) in vv_list[:valid_value_limit]:
                    try:
                        v_str = str(v)
                        if len(v_str) <= 60:
                            print(f"\t{v_str:60}\tFreq = {freq:,}", file=out)
                        else:
                            print(f"\t{v_str[:60]}\tFreq = {freq:,} (Value truncated actual width = {len(v_str)} chars)", file=out)
                    except Exception as e:
                        print(e, file=out)
                        print("Freq = {:,}".format(freq), file=out)
                if len(vv) > valid_value_limit:
                    print("\t--More values not printed--", file=out)
            else:
                print("{} (col {}) had no data values".format(col, col_cnt), file=out)

            print('', file=out)
            if col in self.column_data_types_counts:
                type_counts = self.column_data_types_counts[col]
                print("\tApparent Data Types:", file=out)
                for row_type, freq in sorted(list(type_counts.items()), key=itemgetter(1), reverse=True):
                    print(f'\t\t{freq:,} rows appear to be {row_type}', file=out)
